from jsonclasses.cgraph import CGraph
from jsonclasses.uconf import uconf
from .id_query import id_query
from .class_api import class_api
from .data_requests_and_client import data_requests_and_clients
from .links_interface import links_interface
from .request_manager import request_manager
from .session_manager import session_manager
from .session_input import session_input
from .session import session
from .session_items import session_items
from .data_interface import data_interface
from .string_query import string_query
from .number_query import number_query
from .boolean_query import boolean_query
from .date_query import date_query
from .data_enum import data_enum
from ...utils.join_lines import join_lines
from ...utils.package_utils import session_input_cdefs


def main_program_content(cgraph: CGraph) -> str:
    session_classes = session_items(cgraph)
    use_session = len(session_classes) > 0
    request_url = uconf()['package.ts.url']
    return join_lines([
        _import_lines(),
        *map(lambda e: data_enum(e), cgraph._enum_map.values()),
        string_query(),
        number_query(),
        boolean_query(),
        date_query(),
        id_query(),
        links_interface(),
        *map(lambda c: data_interface(c), cgraph._map.values()),
        *map(lambda c: session_input(c), session_input_cdefs(cgraph)),
        session(session_classes),
        session_manager(session_classes) if use_session else '',
        request_manager(request_url,use_session),
        *map(lambda c: data_requests_and_clients(c), cgraph._map.values()),
        class_api(cgraph,use_session),
        _export_api()
    ], 3)


def _import_lines() -> str:
    return """
import axios from 'axios'
import { stringify } from 'qsparser-js'
    """.strip()

def _export_api() -> str:
    return "export const api = new API()"



from os import getcwd
from pathlib import Path
from click import group, argument, option, echo
from .new import new as execute_new
from .upgrade import upgrade as execute_upgrade
from .package import package as execute_package
from .console import console as execute_console
from .version import version


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    echo(f'JSONClasses CLI {version}')
    ctx.exit()


@group()
@option('-v', '--version', is_flag=True, callback=print_version, expose_value=False, is_eager=True, help='Show version and exit.')
def app():
    pass


@app.command(help='Upgrade this CLI to the newest version.')
def upgrade():
    execute_upgrade()


@app.command(help='Create a new project.')
@argument('name')
@option('-i', '--interactive/--no-interactive', ' /-I', default=True, help='Whether run interactively.')
@option('-u', '--include-user', is_flag=True, required=False, default=None, help='Whether include user.')
@option('-a', '--include-admin', is_flag=True, required=False, default=None, help='Whether include admin.')
@option('-g', '--git-init/--no-git-init', ' /-G', is_flag=True, required=False, default=None, help='Whether create a git repo.')
@option('-m', '--venv/--no-venv', ' /-M', is_flag=True, required=False, default=None, help='Whether create a venv.')
def new(name: str,
        interactive: bool | None,
        include_user: bool | None,
        include_admin: bool | None,
        git_init: bool | None,
        venv: bool | None):
    execute_new(Path(getcwd()) / name,
                interactive=interactive,
                include_user=include_user,
                include_admin=include_admin,
                git_init=git_init,
                venv=venv)


@app.command(help='Launch an interactive console.')
@argument('file', default='app.py')
def console(file: str | None):
    dest = Path(getcwd())
    app_file = dest / file
    execute_console(dest, app_file)


@app.command(help='Generate a client package.')
@argument('lang')
@argument('file', default='app.py')
def package(lang: str, file: str | None):
    dest = Path(getcwd())
    app_file = dest / file
    execute_package(dest, app_file, lang)


if __name__ == '__main__':
    app()



from .codable_struct import codable_struct, codable_struct_item


def session(items: dict[str, str]) -> str:
    struct_items: list[str] = []
    optional = len(items) != 1
    for (s, c) in items.items():
        struct_item = codable_struct_item('public', 'let', s, c, optional)
        struct_items.append(struct_item)
    return codable_struct('Session', [
        codable_struct_item('public', 'let', 'token', 'String', False),
        *struct_items
    ])



from jsonclasses.cdef import CDef
from .interface import InterfaceItem, interface, interface_item
from .jtype_to_ts_type import jtype_to_ts_type


def session_input(cdef: CDef) -> str:
    items: list[InterfaceItem] = []
    (identities, bys) = _session_input_items(cdef)
    identities_optional = len(identities) != 1
    bys_optional = len(bys) != 1
    for (n, t) in identities.items():
        items.append(interface_item(n, t, identities_optional))
    for (n, t) in bys.items():
        items.append(interface_item(n, t, bys_optional))
    name = cdef.name + 'SessionInput'
    return interface(name, items)


def _session_input_items(cdef: CDef) -> tuple[dict[str, str], dict[str, str]]:
    identities: dict[str, str] = {}
    bys: dict[str, str] = {}
    for field in cdef.fields:
        fname = field.json_name
        ftype = jtype_to_ts_type(field.fdef, 'C')
        if field.fdef.auth_identity:
            identities[fname] = ftype
        elif field.fdef.auth_by:
            bys[field.json_name] = ftype
    return (identities, bys)



from jsonclasses.cdef import CDef
from jsonclasses.fdef import FDef, FStore
from jsonclasses.cgraph import CGraph


def class_needs_api(cdef: CDef) -> bool:
    return hasattr(cdef.cls, 'aconf')


def session_input_cdefs(cgraph: CGraph) -> list[CDef]:
    items: list[CDef]  = []
    for cdef in cgraph._map.values():
        if class_needs_session(cdef):
            items.append(cdef)
    return items


def class_needs_session(cdef: CDef) -> bool:
    return hasattr(cdef.cls, 'auth_conf')


def to_create_input(cdef: CDef) -> str:
    return cdef.name + 'CreateInput'


def to_update_input(cdef: CDef) -> str:
    return cdef.name + 'UpdateInput'


def to_create_request(cdef: CDef) -> str:
    return cdef.name + 'CreateRequest'


def to_update_request(cdef: CDef) -> str:
    return cdef.name + 'UpdateRequest'


def to_delete_request(cdef: CDef) -> str:
    return cdef.name + 'DeleteRequest'


def to_delete_many_request(cdef: CDef) -> str:
    return cdef.name + 'DeleteManyRequest'


def to_update_many_request(cdef: CDef) -> str:
    return cdef.name + 'UpdateManyRequest'


def to_upsert_request(cdef: CDef) -> str:
    return cdef.name + 'UpsertRequest'


def to_create_many_request(cdef: CDef) -> str:
    return cdef.name + 'CreateManyRequest'


def to_id_request(cdef: CDef) -> str:
    return cdef.name + 'IDRequest'


def to_list_request(cdef: CDef) -> str:
    return cdef.name + 'ListRequest'


def to_single_query(cdef: CDef) -> str:
    return cdef.name + 'SingleQuery'


def to_list_query(cdef: CDef) -> str:
    return cdef.name + 'ListQuery'


def to_seek_query(cdef: CDef) -> str:
    return cdef.name + 'SeekQuery'


def to_query_data(cdef: CDef) -> str:
    return cdef.name + 'QueryData'


def to_result(cdef: CDef) -> str:
    return cdef.name


def to_list_result(cdef: CDef, mode: str = 'swift') -> str:
    return '[' + cdef.name + ']'


def to_result_picks(cdef: CDef) -> str:
    return cdef.name + 'ResultPick'


def to_include(cdef: CDef) -> str:
    return cdef.name + 'Include'


def to_include_key(cname: str, name: str) -> str:
    return cname + name.capitalize() + 'Include'


def to_sort_orders(cdef: CDef) -> str:
    return cdef.name + 'SortOrder'


def to_session_input(cdef: CDef) -> str:
    return cdef.name + 'SessionInput'


def to_session(cdef: CDef, mode: str = 'ts') -> str:
    if mode == 'swift':
        return 'Session'
    return cdef.name + 'Session'


def to_client(cdef: CDef) -> str:
    return cdef.name + 'Client'


def to_sign_in_request(cdef: CDef) -> str:
    return cdef.name + 'SignInRequest'


def is_field_link(fdef: FDef) -> bool:
    return fdef.fstore == FStore.LOCAL_KEY or fdef.use_join_table



def date_query() -> str:
    return """
interface DateValueQuery {
    _gt?: Date
    _gte?: Date
    _lt?: Date
    _lte?: Date
    _on?: Date
}

interface DateEqQuery {
    _eq: Date
}

interface DateNeqQuery {
    _neq: Date
}

interface DateNullQuery {
    _null: boolean
}

interface DateOrQuery {
    _or: DateQuery[]
}

interface DateAndQuery {
    _and: DateQuery[]
}

interface DateBeforeQuery {
    _before: Date
}

interface DateAfterQuery {
    _after: Date
}

export type DateQuery = Date | DateValueQuery | DateEqQuery | DateNeqQuery | DateNullQuery | DateOrQuery | DateAndQuery |
                        DateBeforeQuery | DateAfterQuery
    """.strip() + "\n"



from ...utils.join_lines import join_lines


def interface_enum(name: str, items: list[str]) -> str:
    lines = [_interface_enum_first_line(name), *items, "}"]
    return join_lines(lines, 1)


def _interface_enum_first_line(name: str) -> str:
    return f"enum {name} {'{'}"


def interface_enum_item(name: str, value: str) -> str:
    return f"    {name} = '{value}',"



def string_query() -> str:
    return """
type Mode = 'default' | 'insensitive'

interface StringContainsQuery {
    _contains: string
    _mode?: Mode
}

interface StringPrefixQuery {
    _prefix: string
    _mode?: Mode
}

interface StringSuffixQuery {
    _suffix: string
    _mode?: Mode
}

interface StringMatchQuery {
    _match: string
    _mode?: Mode
}

interface StringEqQuery {
    _eq: string
}

interface StringNeqQuery {
    _neq: string
}

interface StringNullQuery {
    _null: boolean
}

interface StringCompareQuery {
    _gt?: string
    _gte?: string
    _lt?: string
    _lte?: string
}

interface StringOrQuery {
    _or: StringQuery[]
}

interface StringAndQuery {
    _and: StringQuery[]
}

export type StringQuery = string | StringContainsQuery | StringPrefixQuery | StringSuffixQuery | StringMatchQuery |
                          StringEqQuery | StringNeqQuery | StringNullQuery | StringCompareQuery | StringOrQuery | StringAndQuery
    """.strip() + "\n"



from typing import cast
from jsonclasses.cdef import CDef
from jsonclasses_server.aconf import AConf
from ...utils.package_utils import (class_needs_session, to_include, to_result_picks,
                                    to_session_input, to_sign_in_request, to_session, to_single_query)
from .shared_utils import interface_required_include

def sign_in_request(cdef: CDef) -> str:
    if not class_needs_session(cdef):
        return ''
    name = cast(AConf, cdef.cls.aconf).name
    return f"""
class {to_sign_in_request(cdef)}<T extends Partial<{to_session(cdef)}>> extends Promise<T> {'{'}

    #input: {to_session_input(cdef)}
    #query?: {to_single_query(cdef)}

    constructor(input: {to_session_input(cdef)}, query?:{to_single_query(cdef)}){'{'}
        super(() => {'{'}{'}'})
        this.#input = input
        this.#query = query
    {'}'}

    {_data_query_request_common(cdef, to_sign_in_request(cdef))}
    {_data_query_request_includes(cdef, to_sign_in_request(cdef)) if interface_required_include(cdef) else ''}
    async exec(): Promise<{to_session(cdef)}> {'{'}
        const session = await RequestManager.share.post('/{name}/session', this.#input, this.#query) as {to_session(cdef)}
        SessionManager.share.setSession(session)
        return session
    {'}'}
{'}'}
    """.strip() + "\n"


def _data_query_request_includes(cdef: CDef, request: str) -> str:
    return f"""
    include(includes: {to_include(cdef)}[]): {request}<T> {'{'}
        this.#query = {'{'}...this.#query, _includes: includes {'}'}
        return this
    {'}'}
        """.strip() + "\n"


def _data_query_request_common(cdef: CDef, request: str) -> str:
    return f"""
    pick(picks: {to_result_picks(cdef)}[]): {request}<T> {'{'}
        this.#query = {'{'}...this.#query, _pick: picks{'}'}
        return this
    {'}'}

    omit(omits: {to_result_picks(cdef)}[]): {request}<T> {'{'}
        this.#query = {'{'}...this.#query, _omit: omits{'}'}
        return this
    {'}'}
    """.strip() + "\n"



from typing import Literal
from ...utils.join_lines import join_lines


def codable_struct(name: str, items: list[str]) -> str:
    lines = [_codable_struct_first_line(name), *items, "}"]
    return join_lines(lines, 1)


def codable_struct_class(name: str, items: list[str]) -> str:
    lines = [_codable_struct_first_line(name, False), *items, "}"]
    return join_lines(lines, 1)


def _codable_struct_first_line(name: str, struct: bool = True) -> str:
    return f"public {'struct' if struct else 'class'} {name}: Codable {'{'}"


def codable_struct_item(access: Literal["public", "fileprivate", "private"],
                        mutability: Literal["let", "var"],
                        name: str,
                        field_type: str,
                        optional: bool,
                        default: str | None = None) -> str:
    return (f"    {access} {mutability} {name}: {field_type}"
            f"{'?' if optional else ''}{f' = {default}' if default else ''}")



from typing import Literal


def user_content() -> str:
    return "\n\n" + """
@authorized
@api
@pymongo
@jsonclass
class User:
    id: str = types.readonly.str.primary.mongoid.required
    email: str = types.str.email.unique.authidentity.required
    password: str = types.writeonly.str.securepw.length(8, 16).salt.authbycheckpw.required
    name: str | None
    created_at: datetime = types.readonly.datetime.tscreated.required
    updated_at: datetime = types.readonly.datetime.tsupdated.required
    """.strip() + "\n"


def admin_content() -> str:
    return "\n\n" + """
@authorized
@api
@pymongo
@jsonclass
class Admin:
    id: str = types.readonly.str.primary.mongoid.required
    email: str = types.str.email.unique.authidentity.required
    password: str = types.writeonly.str.securepw.length(8, 16).salt.authbycheckpw.required
    name: str | None
    created_at: datetime = types.readonly.datetime.tscreated.required
    updated_at: datetime = types.readonly.datetime.tsupdated.required
    """.strip() + "\n"


def app_content(include_user: bool, include_admin: bool) -> str:

    return f"""
from __future__ import annotations
from datetime import datetime
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo
from jsonclasses_server import api, {'authorized, ' if include_user or include_admin else ''}server




# You can create a model by using class and type hint syntax, with decorators
# decorated.
#
# Models with relationships are supported.
#
# To understand the types modifier pipeline syntax, check our documentation:
#
# https://docs.jsonclasses.com/docs/api-documentation/types-modifiers
#
# @api
# @pymongo
# @jsonclass
# class MyModel:
#     id: str = types.readonly.str.primary.mongoid.required
#     one: str
#     two: int = 1
#     created_at: datetime = types.readonly.datetime.tscreated.required
#     updated_at: datetime = types.readonly.datetime.tsupdated.required
{user_content() if include_user else ''}{admin_content() if include_admin else ''}

app = server()


# You can still write custom routes if synthesized routes don't satisfy your
# need. To understand the ORM methods, check our documentation:
#
# https://docs.jsonclasses.com/docs/api-documentation/orm-addons
    """.strip() + '\n'



def req_content(include_user: bool, include_admin: bool) -> str:
    nl = '\n'
    return f"""
jsonclasses>=3.2.0,<4.0.0
jsonclasses-pymongo>=3.2.0,<4.0.0
jsonclasses-server>=3.2.0,<4.0.0
thunderlight>=0.1.2,<1.0.0
inflection-plus>=0.1.0,<1.0.0{f'{nl}bcrypt>=3.2.0,<4.0.0' if include_admin or include_user else ''}
    """.strip() + '\n'



from typing import cast
from jsonclasses.cdef import CDef
from .sign_in_request import sign_in_request
from jsonclasses_server.aconf import AConf
from .shared_utils import interface_required_include
from ...utils.join_lines import join_lines
from ...utils.package_utils import (
    class_needs_api, class_needs_session, to_client, to_create_input, to_create_request, to_delete_request, to_id_request,
    to_list_query, to_list_request, to_query_data, to_result, to_result_picks, to_seek_query, to_session, to_session_input,
    to_sign_in_request, to_single_query, to_update_input, to_update_request, to_sort_orders, to_include, to_upsert_request,
    to_create_many_request, to_update_many_request, to_delete_many_request
)


def data_requests_and_clients(cdef: CDef) -> str:
    if not class_needs_api(cdef):
        return ''
    aconf = cast(AConf, cdef.cls.aconf)
    return join_lines([
        _data_create_requet(cdef, aconf.name) if 'C' in aconf.actions else '',
        _data_update_request(cdef, aconf.name) if 'U' in aconf.actions else '',
        _data_delete_request(cdef, aconf.name) if 'D' in aconf.actions else '',
        _data_id_request(cdef, aconf.name) if 'R' in aconf.actions else '',
        _data_upsert_request(cdef, aconf.name) if all(element in aconf.actions for element in ['C','U']) else '',
        _data_create_many_request(cdef, aconf.name) if 'C' in aconf.actions else '',
        _data_update_many_request(cdef, aconf.name) if 'U' in aconf.actions else '',
        _data_delete_many_request(cdef, aconf.name) if 'D' in aconf.actions else '',
        _data_list_request(cdef, aconf.name) if 'L' in aconf.actions else '',
        sign_in_request(cdef),
        _data_client(cdef, aconf)
    ], 2)


def _data_query_request_common(cdef: CDef, request: str) -> str:
    return f"""
    pick(picks: {to_result_picks(cdef)}[]): {request}<Pick<T, typeof picks[number]>> {'{'}
        this.#query = {'{'}...this.#query, _pick: picks{'}'}
        return this
    {'}'}

    omit(omits: {to_result_picks(cdef)}[]): {request}<Omit<T, typeof omits[number]>> {'{'}
        this.#query = {'{'}...this.#query, _omit: omits{'}'}
        return this
    {'}'}
    """.strip() + "\n"


def _data_query_request_includes(cdef: CDef, request: str) -> str:
    if interface_required_include(cdef):
        return f"""
    include(includes: {to_include(cdef)}[]): {request}<T> {'{'}
        this.#query = {'{'}...this.#query, _includes: includes {'}'}
        return this
    {'}'}
        """.strip() + "\n"

    return ''


def _data_create_requet(cdef:CDef, name:str) -> str:
    return f"""
class {to_create_request(cdef)}<T extends Partial<{to_result(cdef)}>> extends Promise<T> {'{'}

    #input: {to_create_input(cdef)}
    #query?: {to_single_query(cdef)}

    constructor(input: {to_create_input(cdef)}, query?:{to_single_query(cdef)}){'{'}
        super(() => {'{'}{'}'})
        this.#input = input
        this.#query = query
    {'}'}

    {_data_query_request_common(cdef,to_create_request(cdef))}
    {_data_query_request_includes(cdef, to_create_request(cdef))}
    async exec(): Promise<T> {'{'}
        return await RequestManager.share.post('/{name}', this.#input, this.#query)
    {'}'}
{'}'}
    """.strip() + "\n"


def _data_upsert_request(cdef:CDef, name:str) -> str:
    return  f"""
class {to_upsert_request(cdef)}<T extends Partial<{to_result(cdef)}>> extends Promise<T> {'{'}

    #input: {to_query_data(cdef)}

    constructor(input: {to_query_data(cdef)}){'{'}
        super(() => {'{'}{'}'})
        this.#input = input
    {'}'}

    async exec(): Promise<T> {'{'}
        return await RequestManager.share.post('/{name}', {'{'} '_upsert': this.#input {'}'})
    {'}'}
{'}'}
    """


def _data_create_many_request(cdef: CDef, name:str) -> str:
    return f"""
class {to_create_many_request(cdef)}<T extends Partial<{to_result(cdef)}>> extends Promise<T> {'{'}

    #input: {to_create_input(cdef)}[]
    #query?: {to_single_query(cdef)}

     constructor(input: {to_create_input(cdef)}[], query?:{to_single_query(cdef)}){'{'}
        super(() => {'{'}{'}'})
        this.#input = input
        this.#query = query
    {'}'}

    {_data_query_request_common(cdef, to_create_many_request(cdef))}
    {_data_query_request_includes(cdef, to_create_many_request(cdef))}
    async exec(): Promise<T[]> {'{'}
        return await RequestManager.share.post('/{name}', {'{'} '_create': this.#input {'}'})
    {'}'}
{'}'}
    """


def _data_update_many_request(cdef:CDef, name:str) -> str:
    return f"""
class {to_update_many_request(cdef)}<T extends Partial<{to_result(cdef)}>> extends Promise<T> {'{'}

    #input: {to_query_data(cdef)}
    #query?: {to_single_query(cdef)}

    constructor(input: {to_query_data(cdef)}, query?:{to_single_query(cdef)}) {'{'}
        super(() => {'{'}{'}'})
        this.#input = input
        this.#query = query
    {'}'}

    {_data_query_request_common(cdef, to_update_many_request(cdef))}
    {_data_query_request_includes(cdef, to_update_many_request(cdef))}
    async exec(): Promise<{to_result(cdef)}> {'{'}
        return await RequestManager.share.patch('/{name}', {'{'} '_update': this.#input {'}'})
    {'}'}
{'}'}
    """


def _data_update_request(cdef:CDef, name:str) -> str:
    return f"""
class {to_update_request(cdef)}<T extends Partial<{to_result(cdef)}>> extends Promise<T> {'{'}

    #id: string
    #input: {to_update_input(cdef)}
    #query?: {to_single_query(cdef)}

    constructor(id:string, input: {to_update_input(cdef)}, query?: {to_single_query(cdef)},) {'{'}
        super(() => {'{'}{'}'})
        this.#id = id
        this.#input = input
        this.#query = query
    {'}'}

    {_data_query_request_common(cdef, to_update_request(cdef))}
    {_data_query_request_includes(cdef, to_update_request(cdef))}
    async exec(): Promise<{to_result(cdef)}> {'{'}
        return await RequestManager.share.patch(`/{name}/${'{'}this.#id{'}'}`, this.#input, this.#query)
    {'}'}
{'}'}
""".strip() + "\n"


def _data_delete_request(cdef:CDef, name:str) -> str:
    return f"""
class {to_delete_request(cdef)} extends Promise<void> {'{'}

    #id: string

    constructor(id: string) {'{'}
        super(() => {'{'}{'}'})
        this.#id = id
    {'}'}
    async exec(): Promise<void> {'{'}
        return await RequestManager.share.delete(`/{name}/${'{'}this.#id{'}'}`)
    {'}'}
{'}'}
""".strip() + "\n"


def _data_delete_many_request(cdef:CDef, name:str) -> str:
    return f"""
class {to_delete_many_request(cdef)} extends Promise<void> {'{'}

    #query?: {to_seek_query(cdef)}

    constructor(query?: {to_seek_query(cdef)}) {'{'}
        super(() => {'{'}{'}'})
        this.#query = query
    {'}'}

    async exec(): Promise<void> {'{'}
        return await RequestManager.share.delete('/{name}', this.#query)
    {'}'}
{'}'}
    """


def _data_id_request(cdef:CDef, name:str) -> str:
    return f"""
class {to_id_request(cdef)}<T extends Partial<{to_result(cdef)}>> extends Promise<T> {'{'}

    #id: string
    #query?: {to_single_query(cdef)}

    constructor(id: string, query?: {to_single_query(cdef)}) {'{'}
        super(() => {'{'}{'}'})
        this.#id = id,
        this.#query = query
    {'}'}

    {_data_query_request_common(cdef, to_id_request(cdef))}
    { _data_query_request_includes(cdef, to_id_request(cdef))}
    async exec(): Promise<{to_result(cdef)}> {'{'}
        return await RequestManager.share.get(`/{name}/${'{'}this.#id{'}'}`, this.#query)
    {'}'}
{'}'}
""".strip() + "\n"


def _data_list_request(cdef:CDef, name:str) -> str:
    return f"""
class {to_list_request(cdef)}<T extends Partial<{to_result(cdef)}>> extends Promise<T[]> {'{'}

    #query?: {to_list_query(cdef)}

    constructor(query?: {to_list_query(cdef)}) {'{'}
        super(() => {'{'}{'}'})
        this.#query = query
    {'}'}

    order(order: {to_sort_orders(cdef)} | {to_sort_orders(cdef)}[]): {to_list_request(cdef)}<T> {'{'}
        this.#query = {'{'}...this.#query, _order: order{'}'}
        return this
    {'}'}

    skip(skip: number): {to_list_request(cdef)}<T> {'{'}
        this.#query = {'{'}...this.#query, _skip: skip{'}'}
        return this
    {'}'}

    limt(limit: number): {to_list_request(cdef)}<T> {'{'}
        this.#query = {'{'}...this.#query, _limit:limit{'}'}
        return this
    {'}'}

    pageSize(pageSize: number): {to_list_request(cdef)}<T> {'{'}
        this.#query = {'{'}...this.#query, _pageSize: pageSize{'}'}
        return this
    {'}'}

    pageNo(pageNo: number): {to_list_request(cdef)}<T> {'{'}
        this.#query = {'{'}...this.#query, _pageNo: pageNo{'}'}
        return this
    {'}'}

    {_data_query_request_common(cdef, to_list_request(cdef))}
    {_data_query_request_includes(cdef, to_list_request(cdef))}
    async exec(): Promise<{to_result(cdef)}[]> {'{'}
        return await RequestManager.share.get('/{name}',this.#query)
    {'}'}
{'}'}
""".strip() + "\n"


def _data_client(cdef: CDef, aconf: AConf) -> str:
    return join_lines([
        f'class {to_client(cdef)} {"{"}',
        _data_client_create(cdef, aconf),
        _data_client_id(cdef, aconf),
        _data_client_update(cdef, aconf),
        _data_client_upsert(cdef, aconf),
        _data_client_find(cdef, aconf),
        _data_client_delete(cdef, aconf),
        _sign_in(cdef),
        '}'
    ], 2)


def _data_client_create(cdef: CDef, aconf: AConf) -> str:
    if 'C' not in aconf.actions:
        return ''
    return join_lines([
        f'    create(input: {to_create_input(cdef)}, query?: {to_single_query(cdef)}): {to_create_request(cdef)}<{cdef.name}> {"{"}',
        f'        return new {to_create_request(cdef)}(input, query)',
        '    }',
        '\n',
        f'    createMany(input: {to_create_input(cdef)}[]): {to_create_many_request(cdef)}<{cdef.name}> {"{"}',
        f'        return new {to_create_many_request(cdef)}(input)',
        '    }',
    ])


def _data_client_id(cdef: CDef, aconf: AConf) -> str:
    if 'R' not in aconf.actions:
        return ''
    return join_lines([
        f'    id(id: string, query?: {to_single_query(cdef)}) {"{"}',
        f'        return new {to_id_request(cdef)}(id, query)',
        '    }',
    ])


def _data_client_update(cdef: CDef, aconf: AConf) -> str:
    if 'U' not in aconf.actions:
        return ''
    return join_lines([
        f'    update(id: string, input: {to_update_input(cdef)}, query?: {to_single_query(cdef)}): {to_update_request(cdef)}<{cdef.name}> {"{"}',
        f'        return new {to_update_request(cdef)}(id, input, query)',
        '    }',
        '\n',
        f'    updateMany(input: {to_query_data(cdef)}): {to_update_many_request(cdef)}<{cdef.name}> {"{"}',
        f'        return new {to_update_many_request(cdef)}(input)',
        '    }'
    ])


def _data_client_upsert(cdef: CDef, aconf: AConf) -> str:
    if not all( element in aconf.actions for element in ['C','U']):
        return ''
    return join_lines([
        f'    upsert(input: {to_query_data(cdef)}): {to_upsert_request(cdef)}<{cdef.name}> {"{"}',
        f'        return new {to_upsert_request(cdef)}(input)',
        '    }',
    ])



def _data_client_find(cdef: CDef, aconf: AConf) -> str:
    if 'L' not in aconf.actions:
        return ''
    return join_lines([
        f'    find(query?: {to_list_query(cdef)}): {to_list_request(cdef)}<{cdef.name}> {"{"}',
        f'        return new {to_list_request(cdef)}(query)',
        '    }',
    ])


def _data_client_delete(cdef: CDef, aconf: AConf) -> str:
    if 'D' not in aconf.actions:
        return ''
    return join_lines([
        f'    delete(id: string): {to_delete_request(cdef)} {"{"}',
        f'        return new {to_delete_request(cdef)}(id)',
        '    }',
        '\n',
        f'    deleteMany(query?: {to_seek_query(cdef)}): {to_delete_many_request(cdef)} {"{"}',
        f'        return new {to_delete_many_request(cdef)}(query)',
        '    }',
    ])


def _sign_in(cdef: CDef) -> str:
    if not class_needs_session(cdef):
        return ''
    return join_lines([
        '\n',
        f'    signIn(input: {to_session_input(cdef)}, query?: {to_single_query(cdef)}): {to_sign_in_request(cdef)}<{to_session(cdef)}>{"{"}',
        f'       return new {to_sign_in_request(cdef)}(input, query)',
        '    }'
    ])



from jsonclasses.cdef import CDef
from jsonclasses_cli.utils.package_utils import to_session_input
from .codable_struct import codable_struct, codable_struct_item
from .jtype_to_swift_type import jtype_to_swift_type


def session_input(cdef: CDef) -> str:
    name = to_session_input(cdef)
    struct_items: list[str] = []
    (identities, bys) = _session_input_items(cdef)
    identities_optional = len(identities) != 1
    bys_optional = len(bys) != 1
    for (n, t) in identities.items():
        struct_item = codable_struct_item('public', 'let', n, t, identities_optional)
        struct_items.append(struct_item)
    for (n, t) in bys.items():
        struct_item = codable_struct_item('public', 'let', n, t, bys_optional)
        struct_items.append(struct_item)
    return codable_struct(name, struct_items)


def _session_input_items(cdef: CDef) -> tuple[dict[str, str], dict[str, str]]:
    identities: dict[str, str] = {}
    bys: dict[str, str] = {}
    for field in cdef.fields:
        fname = field.json_name
        ftype = jtype_to_swift_type(field.fdef, 'C')
        if field.fdef.auth_identity:
            identities[fname] = ftype
        elif field.fdef.auth_by:
            bys[field.json_name] = ftype
    return (identities, bys)


from jsonclasses.cgraph import CGraph
from jsonclasses.uconf import uconf
from jsonclasses_cli.package.swift.link_codable import link_codable
from .session_input import session_input
from .import_lines import import_lines
from .string_query import string_query
from .int_query import int_query
from .float_query import float_query
from .bool_query import bool_query
from .date_query import date_query
from .id_query import id_query
from .sort_order import sort_order
from .data_enum import data_enum
from .data_class import data_class
from .session_items import session_items
from .session import session
from .response import response_struct
from .user_default import user_default
from .session_manager import session_manager
from .sign_out import sign_out
from .request_manager import request_manager
from .data_requests_and_clients import data_requests_and_clients, data_client_instances
from ...utils.join_lines import join_lines
from ...utils.package_utils import session_input_cdefs


def main_program_content(cgraph: CGraph) -> str:
    session_classes = session_items(cgraph)
    use_session = len(session_classes) > 0
    request_url = uconf()['package.swift.url']
    return join_lines([
        import_lines(),
        string_query(),
        int_query(),
        float_query(),
        bool_query(),
        date_query(),
        id_query(),
        # list query
        # dict query
        sort_order(),
        link_codable(),
        *map(lambda e: data_enum(e), cgraph._enum_map.values()),
        *map(lambda c: data_class(c), cgraph._map.values()),
        *map(lambda c: session_input(c), session_input_cdefs(cgraph)),
        session(session_classes) if use_session else '',
        response_struct(),
        user_default(),
        session_manager() if use_session else '',
        sign_out() if use_session else '',
        request_manager(request_url, use_session),
        *map(lambda c: data_requests_and_clients(c), cgraph._map.values()),
        join_lines(map(lambda c: data_client_instances(c), cgraph._map.values()), 1),
    ], 2)



from inflection import camelize
from jsonclasses.cdef import CDef
from jsonclasses.fdef import FType
from .unary_sort_order import unary_sort_order
from .codable_struct import codable_struct, codable_struct_class, codable_struct_item
from .codable_enum import codable_associated_item, codable_enum, codable_enum_item
from .codable_class import codable_class, codable_class_item
from .jtype_to_swift_type import jtype_to_swift_type
from .shared_utils import (
    class_create_input_items, class_update_input_items, is_field_primary,
    is_field_local_key, field_can_read, field_ref_id_name, is_field_queryable,
    is_field_ref, is_field_required_for_read, array, list_query_items,
    class_include_items, to_many_request_type
)
from ...utils.join_lines import join_lines
from ...utils.package_utils import (
    to_create_input, to_include_key, to_query_data, to_seek_query, to_update_input,
    to_single_query, to_list_query, to_result, to_result_picks, to_include, to_sort_orders
)


def data_class(cdef: CDef) -> str:
    return join_lines([
        _class_create_input(cdef),
        _class_update_input(cdef),
        _class_sort_orders(cdef),
        _class_result_picks(cdef),
        _class_include_key_enums(cdef),
        _class_include_enum(cdef),
        _class_many_request_enum(cdef),
        _class_single_query(cdef),
        _class_seek_query(cdef),
        _class_query_data(cdef),
        _class_list_query(cdef),
        _class_result(cdef),
    ], 2)


def _class_create_input(cdef: CDef) -> str:
    return codable_class(to_create_input(cdef), class_create_input_items(cdef))


def _class_update_input(cdef: CDef) -> str:
    return codable_class(to_update_input(cdef), class_update_input_items(cdef))


def _class_sort_orders(cdef: CDef) -> str:
    fnames: list[str] = []
    for field in cdef.fields:
        if not is_field_queryable(field):
            continue
        if is_field_primary(field):
            continue
        if is_field_ref(field):
            continue
        if not field_can_read(field):
            continue
        fnames.append(camelize(field.name))
    enum_items: list[str] = []
    for name in fnames:
        enum_items.append(codable_enum_item(name, 'String', name))
        desc_name = name + 'Desc'
        enum_items.append(codable_enum_item(desc_name, 'String', "-" + name))
    name = to_sort_orders(cdef)
    enum = codable_enum(name, 'String', enum_items)
    unary = unary_sort_order(name)
    return join_lines([enum, unary], 2)


def _class_result_picks(cdef: CDef) -> str:
    items: list[str] = []
    for field in cdef.fields:
        if not field_can_read(field):
            continue
        name = camelize(field.name)
        items.append(codable_enum_item(name, 'String', name))
        if is_field_local_key(field):
            idname = field_ref_id_name(field)
            items.append(codable_enum_item(idname, 'String', idname))
    return codable_enum(to_result_picks(cdef), 'String', items)


def _class_include_key_enums(cdef: CDef) -> str:
    cname = cdef.name
    enums: list[str] = []
    for field in cdef.fields:
        if is_field_ref(field):
            enums.append(_class_include_key_enum(cname, field.name))
    return join_lines(enums, 2)


def _class_include_key_enum(cname: str, fname: str) -> str:
    return codable_enum(to_include_key(cname, fname), 'Int', [
        codable_enum_item(fname, 'Int', '1')
    ])


def _class_include_enum_init_if(idx: int, key: str, q: str) -> str:
    return f"""
        {'} else ' if idx > 0 else ''}if container.contains(.{key}) {'{'}
            self = .{key}(try! container.decode({q}.self, forKey: .{key}))""".strip('\n')


def _class_include_enum_encode_case(key: str) -> str:
    return f"""
        case .{key}(let value):
            try! container.encode(value, forKey: .{key})""".strip('\n')


def _class_include_enum(cdef: CDef) -> str:
    items = class_include_items(cdef)
    if len(items) == 0:
        return ""
    cases = join_lines(map(lambda i: codable_associated_item(i[0], i[1] + '?'), items), 1)
    coding_keys = join_lines([
        '    enum CodingKeys: String, CodingKey {',
        *map(lambda i: f'        case {i[0]} = "{i[0]}"', items),
        '    }'
    ], 1)
    init = join_lines([
        '    public init(from decoder: Decoder) throws {',
        '        let container = try! decoder.container(keyedBy: CodingKeys.self)',
        join_lines(map(lambda t: _class_include_enum_init_if(t[0], t[1][0], t[1][1]), enumerate(items))),
        '        } else {',
        '            throw NSError()',
        '        }',
        '    }'
    ], 1)
    encode = join_lines([
        '    public func encode(to encoder: Encoder) throws {',
        '        var container = encoder.container(keyedBy: CodingKeys.self)',
        '        switch self {',
        join_lines(map(lambda t: _class_include_enum_encode_case(t[0]), items)),
        '        }',
        '    }'
    ])
    return codable_enum(to_include(cdef), None, [join_lines([
        cases,
        coding_keys,
        init,
        encode
    ], 2)])


def _single_query_items(cdef: CDef) -> list[str]:
    result_picks = array(to_result_picks(cdef))
    result_includes = array(to_include(cdef))
    pick = codable_struct_item(
        'fileprivate', 'var', '_pick', result_picks, True, 'nil')
    omit = codable_struct_item(
        'fileprivate', 'var', '_omit', result_picks, True, 'nil')
    if len(class_include_items(cdef)) == 0:
        includes = ""
    else:
        includes = codable_struct_item(
            'fileprivate', 'var', '_includes', result_includes, True, 'nil')
    return [pick, omit, includes]


def _single_query_picks_omits(cdef: CDef, single: bool = True) -> str:
    rpname = to_result_picks(cdef)
    return f"""
    public static func pick(_ picks: [{rpname}]) -> {to_single_query(cdef) if single else to_list_query(cdef)} {'{'}
        let instance = {to_single_query(cdef) if single else to_list_query(cdef)}()
        instance._pick = picks
        return instance
    {'}'}

    public func pick(_ picks: [{rpname}]) -> {to_single_query(cdef) if single else to_list_query(cdef)} {'{'}
        _pick = picks
        return self
    {'}'}

    public static func omit(_ omits: [{rpname}]) -> {to_single_query(cdef) if single else to_list_query(cdef)} {'{'}
        let instance = {to_single_query(cdef) if single else to_list_query(cdef)}()
        instance._omit = omits
        return instance
    {'}'}

    public func omit(_ omits: [{rpname}]) -> {to_single_query(cdef) if single else to_list_query(cdef)} {'{'}
        _omit = omits
        return self
    {'}'}""".strip('\n')


def _single_query_include(cdef: CDef, key: str, itype: str, qtype: str, single: bool = True) -> str:
    return f"""
    public static func include(_ ref: {itype}, _ query: {qtype}? = nil) -> {to_single_query(cdef) if single else to_list_query(cdef)} {'{'}
        let instance = {to_single_query(cdef) if single else to_list_query(cdef)}()
        instance._includes = [.{key}(query)]
        return instance
    {'}'}

    public func include(_ ref: {itype}, _ query: {qtype}? = nil) -> {to_single_query(cdef) if single else to_list_query(cdef)} {'{'}
        if _includes == nil {'{'} _includes = [] {'}'}
        _includes!.append(.{key}(query))
        return self
    {'}'}""".strip('\n')


def _single_query_includes(cdef: CDef, single: bool = True) -> str:
    items: list[tuple(str, str, str)] = []
    for field in cdef.fields:
        if is_field_ref(field):
            if field.fdef.ftype == FType.LIST:
                items.append((field.name, to_include_key(cdef.name, field.name), to_list_query(field.foreign_cdef)))
            else:
                items.append((field.name, to_include_key(cdef.name, field.name), to_single_query(field.foreign_cdef)))
    return join_lines(map(lambda i: _single_query_include(cdef, i[0], i[1], i[2], single), items), 2)


def _list_query_orders(order: str, cdef: CDef, single: bool = True) -> str:
    return f"""
    public static func order(_ order: {order}) -> {to_single_query(cdef) if single else to_list_query(cdef)} {"{"}
        let instance = {to_single_query(cdef) if single else to_list_query(cdef)}()
        instance._order = [order]
        return instance
    {"}"}

    public static func order(_ orders: [{order}]) -> {to_single_query(cdef) if single else to_list_query(cdef)} {"{"}
        let instance = {to_single_query(cdef) if single else to_list_query(cdef)}()
        instance._order = orders
        return instance
    {"}"}

    public func order(_ order: {order}) -> {to_single_query(cdef) if single else to_list_query(cdef)} {"{"}
        if _order == nil {"{"} _order = [] {"}"}
        _order!.append(order)
        return self
    {"}"}

    public func order(_ orders: [{order}]) -> {to_single_query(cdef) if single else to_list_query(cdef)} {"{"}
        if _order == nil {"{"} _order = [] {"}"}
        _order!.append(contentsOf: orders)
        return self
    {"}"}"""


def _list_query_limit_skip_pn_ps(cdef: CDef) -> str:
    lspp = ["limit", "skip", "pageNo", "pageSize"]
    reslut = []
    for i in lspp:
        reslut.append(f"""
    public static func {i}(_ {i}: Int) -> {to_list_query(cdef)} {"{"}
        let instance = {to_list_query(cdef)}()
        instance._{i} = {i}
        return instance
    {"}"}

    public func {i}(_ {i}: Int) -> {to_list_query(cdef)} {"{"}
        _{i} = {i}
        return self
    {"}"}""".strip('\n'))
    return join_lines(reslut)

def _class_single_query(cdef: CDef) -> str:
    return codable_struct_class(to_single_query(cdef), [join_lines([
        join_lines(_single_query_items(cdef), 1),
        _single_query_picks_omits(cdef, True),
        _single_query_includes(cdef, True)
    ], 2)])


def _list_query_find(cdef: CDef, query_name: str) -> str:
    items = list_query_items(cdef)
    last = len(items) - 1
    arglist = lambda i: f"        {i[1][0]}: {i[1][1]}? = nil{'' if i[0] == last else ','}"
    return join_lines([
        '    public static func `where`(',
        *map(arglist, enumerate(items)),
        f'    ) -> {query_name} {"{"}',
        f'        let instance = {query_name}()',
        *map(lambda i: f"        instance.{i[1][0]} = {i[1][0]}", enumerate(items)),
        '        return instance',
        '    }',
        '\n',
        '    public func `where`(',
        *map(arglist, enumerate(items)),
        f'    ) -> {query_name} {"{"}',
        *map(lambda i: f"        if {i[0]} != nil {'{'} self.{i[0]} = {i[0]} {'}'}", items),
        '        return self',
        '    }'
    ], 1)


def _class_seek_query(cdef: CDef) -> str:
    items = list(map(lambda i: codable_struct_item('public', 'var', i[0], i[1], True, 'nil'), list_query_items(cdef)))
    operators = [
        '\n',
        join_lines([
            _list_query_find(cdef, to_seek_query(cdef))
        ], 2)
    ]
    items.extend(operators)
    return codable_struct_class(to_seek_query(cdef), items)


def _class_query_data(cdef: CDef) -> str:
    items = [
        codable_class_item('fileprivate', 'var', '_query', to_seek_query(cdef), False),
        codable_class_item('fileprivate', 'var', '_data', to_update_input(cdef), False),
    ]
    return codable_class(to_query_data(cdef), items)


def _class_list_query(cdef: CDef) -> str:
    items = list(map(lambda i: codable_struct_item('public', 'var', i[0], i[1], True, 'nil'), list_query_items(cdef)))
    sort_order = to_sort_orders(cdef)
    sort_orders = array(sort_order)
    order = codable_struct_item(
        'fileprivate', 'var', '_order', sort_orders, True, 'nil')
    limit = codable_struct_item(
        'fileprivate', 'var', '_limit', 'Int', True, 'nil')
    skip = codable_struct_item(
        'fileprivate', 'var', '_skip', 'Int', True, 'nil')
    page_no = codable_struct_item(
        'fileprivate', 'var', '_pageNo', 'Int', True, 'nil')
    page_size = codable_struct_item(
        'fileprivate', 'var', '_pageSize', 'Int', True, 'nil')
    operators = [
        order, limit, skip, page_no, page_size, *_single_query_items(cdef),
        '\n',
        join_lines([
            _list_query_find(cdef, to_list_query(cdef)),
            _list_query_orders(sort_order, cdef, False),
            _list_query_limit_skip_pn_ps(cdef),
            _single_query_picks_omits(cdef, False),
            _single_query_includes(cdef, False)
        ], 2)
    ]
    items.extend(operators)
    return codable_struct_class(to_list_query(cdef), items)


def _class_result(cdef: CDef) -> str:
    items: list[str] = []
    for field in cdef.fields:
        if not field_can_read(field):
            continue
        optional = not is_field_required_for_read(field)
        name = camelize(field.name)
        stype = jtype_to_swift_type(field.fdef, 'R')
        local_key = is_field_local_key(field)
        item = codable_class_item('public', 'let', name, stype, optional)
        items.append(item)
        if local_key:
            idname = field_ref_id_name(field)
            item = codable_class_item('public', 'let', idname, 'String', optional)
            items.append(item)
    name = to_result(cdef)
    return codable_class(name, items, True)


def _class_many_request_enum(cdef: CDef) -> str:
    return join_lines([
        f'public enum {to_many_request_type(cdef)}: Codable {"{"}',
        '    case update',
        '    case create',
        '    case upsert',
        '\n',
        f'    func getContent(input: {to_query_data(cdef)}) -> Dictionary<String, {to_query_data(cdef)}> {"{"}',
        '        if self  == .update {',
        '            return ["_update": input]',
        '        }',
        '        else if self == .upsert {',
        '            return ["_upsert": input]',
        '        }',
        f'        return [String: {to_query_data(cdef)}]()',
        '    }',
        '\n',
        f'    func getContent(input: [{to_create_input(cdef)}]) -> Dictionary<String, [{to_create_input(cdef)}]> {"{"}',
        '        if self  == .create {',
        '            return ["_create": input]',
        '        }',
        f'        return [String: [{to_create_input(cdef)}]]()',
        '    }',
        '}',
    ])



from typing import cast
from jsonclasses.cgraph import CGraph
from jsonclasses_server.auth_conf import AuthConf
from ...utils.package_utils import class_needs_session


def session_items(cgraph: CGraph) -> dict[str, str]:
    result: dict[str, str] = {}
    for cdef in cgraph._map.values():
        if not class_needs_session(cdef):
            continue
        conf = cast(AuthConf, cdef.cls.auth_conf)
        singular_name = conf.info.srname
        class_name = cdef.cls.__name__
        result[singular_name] = class_name
    return result



from inflection import camelize
from jsonclasses.cdef import CDef
from jsonclasses.cgraph import CGraph
from jsonclasses.jfield import JField
from jsonclasses.fdef import (
    FStore, FType, Nullability, ReadRule, Queryability, WriteRule
)
from jsonclasses.modifiers.required_modifier import RequiredModifier
from jsonclasses.modifiers.default_modifier import DefaultModifier
from .codable_class import CodableClassItem, codable_class_item
from .jtype_to_swift_type import jtype_to_swift_type
from ...utils.package_utils import is_field_link, to_list_query, to_single_query


def class_include_items(cdef: CDef) -> list[tuple[str, str]]:
    items: list[tuple[str, str]] = []
    for field in cdef.fields:
        if is_field_ref(field):
            if field.fdef.ftype == FType.LIST:
                items.append((field.name, to_list_query(field.foreign_cdef)))
            else:
                items.append((field.name, to_single_query(field.foreign_cdef)))
    return items


def list_query_items(cdef: CDef) -> list[tuple[str, str]]:
    items: list[tuple[str, str]] = []
    for field in cdef.fields:
        if not is_field_queryable(field):
            continue
        name = camelize(field.name)
        type = jtype_to_swift_type(field.fdef, 'Q')
        if is_field_ref(field):
            if not is_field_local_key(field):
                continue
            idname = field_ref_id_name(field)
            items.append((idname, 'IDQuery'))
        else:
            items.append((name, type))
    return items


def is_field_local_key(field: JField) -> bool:
    return field.fdef.fstore == FStore.LOCAL_KEY


def class_update_input_items(cdef: CDef) -> list[CodableClassItem]:
    items: list[CodableClassItem] = []
    for field in cdef.fields:
        if not field_can_update(field):
            continue
        name = camelize(field.name)
        stype = jtype_to_swift_type(field.fdef, 'U', is_field_link(field.fdef))
        local_key = is_field_local_key(field)
        item = codable_class_item('public', 'var', name, stype, True)
        items.append(item)
        if local_key:
            idname = field_ref_id_name(field)
            item = codable_class_item('public', 'var', idname, 'String', True)
            items.append(item)
    return items


def class_create_input_items(cdef: CDef) -> list[CodableClassItem]:
    items: list[CodableClassItem] = []
    for field in cdef.fields:
        if not field_can_create(field):
            continue
        optional = not is_field_required_for_create(field)
        name = camelize(field.name)
        stype = jtype_to_swift_type(field.fdef, 'C', is_field_link(field.fdef))
        local_key = is_field_local_key(field)
        if local_key:
            optional = True
        item = codable_class_item('public', 'var', name, stype, optional)
        items.append(item)
        if local_key:
            idname = field_ref_id_name(field)
            item = codable_class_item('public', 'var', idname, 'String', True)
            items.append(item)
    return items


def is_field_required_for_create(field: JField) -> bool:
    if field_has_default(field):
        return False
    if is_field_nonnull(field):
        return True
    return next((True for v in field.types.modifier.vs if isinstance(v, RequiredModifier)), False)


def field_ref_id_name(field: JField) -> str:
    rkes = field.cdef.jconf.ref_name_strategy
    kes = field.cdef.jconf.input_key_strategy
    return kes(rkes(field))


def field_has_default(field: JField) -> bool:
    if is_field_nonnull(field):
        return True
    return next((True for v in field.types.modifier.vs if isinstance(v, DefaultModifier)), False)


def is_field_nonnull(field: JField) -> bool:
    if field.fdef.ftype == FType.LIST:
        if field.fdef.fstore == FStore.LOCAL_KEY or field.fdef.fstore == FStore.FOREIGN_KEY:
            if field.fdef.collection_nullability == Nullability.NONNULL:
                return True
    return False


def is_field_primary(field: JField) -> bool:
    return field.fdef.primary


def is_field_ref(field: JField) -> bool:
    if field.fdef.fstore == FStore.LOCAL_KEY:
        return True
    if field.fdef.fstore == FStore.FOREIGN_KEY:
        return True
    return False


def is_field_required_for_read(field: JField) -> bool:
    if is_field_nonnull(field):
        return True
    return next((True for v in field.types.modifier.vs if isinstance(v, RequiredModifier)), False)


def is_field_queryable(field: JField) -> bool:
    if field.fdef.read_rule == ReadRule.NO_READ:
        return False
    if field.fdef.fstore == FStore.TEMP:
        return False
    return field.fdef.queryability != Queryability.UNQUERYABLE


def field_can_create(field: JField) -> bool:
    return field.fdef.write_rule != WriteRule.NO_WRITE


def field_can_update(field: JField) -> bool:
    if field.fdef.write_rule == WriteRule.NO_WRITE:
        return False
    if field.fdef.write_rule == WriteRule.WRITE_ONCE:
        if is_field_required_for_create(field):
            return False
    return True


def field_can_read(field: JField) -> bool:
    if field.fdef.read_rule == ReadRule.NO_READ:
        return False
    if field.fdef.fstore == FStore.TEMP:
        return False
    return True


def to_many_request_type(cdef: CDef) -> str:
    return cdef.name + 'ManyRequestType'

def array(val: str) -> str:
    return '[' + val + ']'



def tsconfig_json_content() -> str:
    return """
{
    "compilerOptions": {
        "baseUrl": ".",
        "outDir": "./lib",
        "declaration": true,
        "target": "es2016",
        "module": "commonjs",
        "jsx": "react",
        "lib": ["ESNext", "DOM"],
        "strict": true,
        "esModuleInterop": true,
        "skipLibCheck": true,
        "forceConsistentCasingInFileNames": true,
        "resolveJsonModule": true
    },
    "include": ["src", "index.ts"],
    "exclude": ["node_modules", "**/__tests__/*"]
}
    """.strip() + '\n'



from enum import Enum
from inflection import camelize
from jsonclasses_cli.package.ts.interface_enum import interface_enum, interface_enum_item


def data_enum(enum: type[Enum]) -> str:
    return interface_enum(enum.__name__, map(_data_enum_item, enum))


def _data_enum_item(option: Enum):
    name = camelize(option.name.lower())
    value = option.name
    return interface_enum_item(name, value)



from jsonclasses_cli.utils.join_lines import join_lines
from .interface import InterfaceItem, interface, interface_item


def session(items: dict[str, str]) -> str:
    interface_items: list[str] = []
    for (s, c) in items.items():
        name = c + 'Session'
        item = interface_item(s, c, False)
        interface_items.append(_session_interface(name, item))
    return join_lines(interface_items, 2)


def _session_interface(name, item: InterfaceItem) -> str:
    return interface(name, [interface_item('token', 'string', False), item])



def package_content() -> str:
    return """
// swift-tools-version:5.5

import PackageDescription

let package = Package(
    name: "API",
    platforms: [
        .macOS(.v12),
        .iOS(.v15)
    ],
    products: [
        .library(
            name: "API",
            targets: ["API"]),
    ],
    dependencies: [
        .package(url: "https://github.com/fillmula/qsparser-swift.git", from: "1.0.1")
    ],
    targets: [
        .target(
            name: "API",
            dependencies: [.product(name: "QSParser", package: "qsparser-swift")])
    ]
)
    """.strip() + '\n'



def link_codable() -> str:
    return """
public class Link: Codable {
    public var _add: String
    public init(link: String) {
        self._add = link
    }
}

public class UnLink: Codable {
    public var _del: String
    public init(unLink: String) {
        self._del = unLink
    }
}

public enum CreateOrLink<T: Codable>: Codable{
    case createInput(T)
    case link(Link)
}

public enum UpdateOrLink<T: Codable>: Codable{
    case updateInput(T)
    case link(Link)
    case unLink(UnLink)
}
    """.strip() + '\n'



from typing import Literal
from pathlib import Path
from sys import path
from importlib import import_module
from os.path import splitext
from unittest.mock import patch
from jsonclasses.cgraph import CGraph
from .kotlin import kotlin
from .swift import swift
from .ts import ts



def package(dest: Path, app_file: Path, lang: Literal['ts', 'swift', 'kotlin'], cgraph_name: str = 'default', silent: bool = False):
    path.append(str(app_file.parent))
    import_module(splitext(app_file.name)[0], str(app_file.parent)).__dict__
    cgraph = CGraph(cgraph_name)
    match lang:
        case 'swift':
            swift(dest, cgraph, silent)
        case 'kotlin':
            kotlin(dest, cgraph, silent)
        case 'ts':
            ts(dest, cgraph, silent)



def user_default() -> str:
    return """
@propertyWrapper
public struct UserDefault<T: Codable> {

    private var key: String

    public var wrappedValue: T? {
        didSet {
            if let wrappedValue = wrappedValue {
                let data = try! JSONEncoder().encode(wrappedValue)
                let string = String(data: data, encoding: .utf8)
                UserDefaults.standard.setValue(string, forKey: key)
            } else {
                UserDefaults.standard.removeObject(forKey: key)
            }
        }
    }

    fileprivate init(key: String) {
        self.key = key
        if let string = UserDefaults.standard.value(forKey: key) as? String {
            let data = string.data(using: .utf8)!
            self.wrappedValue = try! JSONDecoder().decode(T.self, from: data)
        } else {
            self.wrappedValue = nil
        }
    }
}
    """.strip() + '\n'



def id_query():
    return """
public enum IDQuery: Codable {
    case eq(_ value: String)
    case neq(_ value: String)
    case null(_ value: Bool)

    public enum CodingKeys: String, CodingKey {
        case eq = "_eq"
        case neq = "_neq"
        case null = "_null"
    }

    public init(from decoder: Decoder) throws {
        let container = try! decoder.container(keyedBy: CodingKeys.self)
        if container.contains(.eq) {
            self = .eq(try! container.decode(String.self, forKey: .eq))
        } else if container.contains(.neq) {
            self = .neq(try! container.decode(String.self, forKey: .neq))
        } else if container.contains(.null) {
            self = .null(try! container.decode(Bool.self, forKey: .null))
        } else {
            throw NSError()
        }
    }

    public func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        switch self {
        case .eq(let value):
            try! container.encode(value, forKey: .eq)
        case .neq(let value):
            try! container.encode(value, forKey: .neq)
        case .null(let value):
            try! container.encode(value, forKey: .null)
        }
    }
}
    """.strip() + "\n"



from ...utils.join_lines import join_lines

def unary_sort_order(name: str):
    return join_lines([
        _unary_sort_order_first_line(name),
        _unary_sort_if_block(name),
        _unary_sort_order_last_line()
    ])


def _unary_sort_order_first_line(name: str) -> str:
    return f'public prefix func -(rhs: {name}) -> {name} {"{"}'

def _unary_sort_if_block(name: str) -> str:
    return join_lines([
        f'    if rhs.rawValue.starts(with: "-") {"{"}',
        f'        return {name}(rawValue: String(rhs.rawValue.dropFirst()))!',
        f'    {"}"} else {"{"}',
        f'        return {name}(rawValue: "-" + rhs.rawValue)!',
        f'    {"}"}'
    ])

def _unary_sort_order_last_line() -> str:
    return '}'



from ...utils.join_lines import join_lines


def codable_enum(name: str, raw_type: str | None, items: list[str]) -> str:
    lines = [_codable_enum_first_line(name, raw_type), *items, "}"]
    return join_lines(lines, 1)


def _codable_enum_first_line(name: str, raw_type: str | None) -> str:
    return f"public enum {name}: {'' if raw_type is None else f'{raw_type}, '}Codable {'{'}"


def codable_enum_item(name: str, raw_type: str, value: str) -> str:
    return f"    case {name} = {_wrap_value(value, raw_type)}"


def codable_associated_item(name: str, valtype: str) -> str:
    return f"    case {name}(_ value: {valtype})"


def _wrap_value(value: str, raw_type: str) -> str:
    if raw_type == 'String':
        return '"' + value + '"'
    else:
        return value



def boolean_query() -> str:
    return """
interface BooleanEqQuery {
    _eq: boolean
}

interface BooleanNeqQuery {
    _neq: boolean
}

interface BooleanNullQuery {
    _null: boolean
}

interface BooleanOrQuery {
    _or: BooleanQuery[]
}

interface BooleanAndQuery {
    _and: BooleanQuery[]
}

export type BooleanQuery = boolean | BooleanEqQuery | BooleanNeqQuery | BooleanNullQuery | BooleanOrQuery | BooleanAndQuery
    """.strip() + "\n"



def request_manager(base_url: str, use_session: bool) -> str:
    return f"""
struct RequestManager {'{'}

    static let shared = RequestManager()

    let baseURL: String = "{base_url}"

    func qs<T: Codable>(_ query: T? = nil) -> String {'{'}
        if let query = query {'{'}
            return "?" + (try! QSEncoder().encode(query))
        {'}'} else {'{'}
            return ""
        {'}'}
    {'}'}

    func url<T: Codable>(url: String, query: T) -> URL {'{'}
        return URL(string: baseURL + url + qs(query))!
    {'}'}

    func request(method: String, url: String) async throws {'{'}
        let _: Int? = try await request(method: method, url: url, input: nil as Int?, query: nil as Int?)
    {'}'}

    func request<U: Codable>(
        method: String,
        url: String,
        query: U? = nil
    ) async throws {'{'}
        let _: Int? = try await request(method: method, url: url, input: nil as Int?, query: query)
    {'}'}

    func request<U: Codable, V: Codable>(
        method: String,
        url: String,
        query: U? = nil
    ) async throws -> V? {'{'}
        return try await request(method: method, url: url, input: nil as Int?, query: query)
    {'}'}

    func request<T: Codable, U: Codable, V: Codable>(
        method: String,
        url: String,
        input: T? = nil,
        query: U? = nil
    ) async throws -> V? {'{'}
        let url = self.url(url: url, query: query)
        var request = URLRequest(url: url)
        request.httpMethod = method
        if let input = input {'{'}
            request.httpBody = try! JSONEncoder().encode(input)
        {'}'}{_session_setter() if use_session else ''}
        let (data, response) = try await URLSession.shared.data(for: request)
        if let response = response as? HTTPURLResponse {'{'}
            if response.statusCode == 200 {'{'}
                let responseObject = try! JSONDecoder().decode(Response<V>.self, from: data)
                return responseObject.data
            {'}'} else {'{'}
                return nil
            {'}'}
        {'}'} else {'{'}
            return nil
        {'}'}
    {'}'}

    func post<T: Codable, U: Codable, V: Codable>(
        url: String,
        input: T,
        query: U? = nil
    ) async throws -> V {'{'}
        return try await request(method: "POST", url: url, input: input, query: query)!
    {'}'}

    func patch<T: Codable, U: Codable, V: Codable>(
        url: String,
        input: T,
        query: U? = nil
    ) async throws -> V {'{'}
        return try await request(method: "PATCH", url: url, input: input, query: query)!
    {'}'}

    func delete(url: String) async throws {'{'}
        try await request(method: "DELETE", url: url)
    {'}'}

    func delete<U: Codable>(url: String, query: U? = nil) async throws {'{'}
        try await request(method: "DELETE", url: url, query: query)
    {'}'}

    func get<U: Codable, V: Codable>(
        url: String,
        query: U? = nil
    ) async throws -> V? {'{'}
        return try await request(method: "GET", url: url, query: query)!
    {'}'}
{'}'}

    """.strip() + '\n'


def _session_setter() -> str:
    return '\n' + """
        if let session = SessionManager.shared.session {
            request.setValue("Bearer \(session.token)", forHTTPHeaderField: "Authorization")
        }""".strip('\n')



from ...utils.join_lines import join_lines


InterfaceItem = tuple[str, str, bool]


def interface(name: str, items: list[InterfaceItem], export_ability: bool = False) -> str:
    return join_lines([
        join_lines([
            interface_first_line(name, export_ability),
            interface_inst_items(items),
            '}'
        ], 1)
    ], 2)


def interface_first_line(name: str, export_ability: bool = False) -> str:
    return f"{'export ' if export_ability else ''}interface {name} {'{'}"


def interface_type_item(name: str, items: list[str]) -> str:
    return f"type {name} = {' | '.join(items)}"


def interface_include_key_item(name: str, include_type: str) -> str:
    return f"    {name}?: {include_type}"


def interface_pick_omit_items(name: str) -> str:
    return join_lines([
        f"    _pick?: {name}[]",
        f"    _omit?: {name}[]",
    ])


def interface_include_item(name: str) -> str:
    return f"    _includes?: {name}[]"


def list_query_order_item(name: str) -> str:
    return f"    _order?: {name} | {name}[]"


def list_query_limit_skip_pn_ps() -> str:
    lspp = ["limit", "skip", "pageNo", "pageSize"]
    reslut = []
    for i in lspp:
        reslut.append(f"    _{i}?: number")
    return join_lines(reslut)


def interface_inst_items(items: list[InterfaceItem]) -> str:
    return join_lines(map(lambda i: f"    {i[0]}{'?' if i[2] else ''}: {i[1]}", items), 1)


def interface_item(name: str,
                   field_type: str,
                   optional: bool,) -> InterfaceItem:
    return (name, field_type, optional)



from pathlib import Path
from code import interact
from sys import path
from os.path import splitext
from importlib import import_module
from jsonclasses.isjsonclass import isjsonclass


def console(dest: Path, app_file: Path):
    path.append(str(dest))
    result = import_module(splitext(app_file.name)[0], str(dest)).__dict__
    locals = {}
    for (k, v) in result.items():
        if isjsonclass(v):
            locals[k] = v
    models = ", ".join(list(locals.keys()))
    interact(banner=f"JSONClasses Console\nAvailable Models: {models}", local=locals)



from typing import cast
from inflection import camelize, pluralize
from jsonclasses.cdef import CDef
from jsonclasses_server.aconf import AConf
from .codable_class import CodableClassItem
from .shared_utils import class_create_input_items, class_include_items, class_update_input_items, list_query_items, to_many_request_type
from ...utils.join_lines import join_lines
from ...utils.package_utils import (
    class_needs_api, class_needs_session, to_client, to_create_input, to_create_many_request, to_create_request, to_delete_many_request, to_delete_request,
    to_id_request, to_include_key, to_list_query, to_list_request, to_list_result, to_query_data, to_result_picks, to_seek_query, to_session, to_session_input, to_sign_in_request, to_single_query,
    to_update_input, to_result, to_update_many_request, to_update_request, to_sort_orders, to_upsert_request
)


def data_client_instances(cdef: CDef) -> str:
    if not class_needs_api(cdef):
        return ''
    var_name = camelize(pluralize(cdef.name))
    return f'public var {var_name} = {to_client(cdef)}()'


def data_requests_and_clients(cdef: CDef) -> str:
    if not class_needs_api(cdef):
        return ''
    aconf = cast(AConf, cdef.cls.aconf)
    return join_lines([
        _data_create_request(cdef, aconf.name) if 'C' in aconf.actions else '',
        _data_update_request(cdef, aconf.name) if 'U' in aconf.actions else '',
        _data_delete_request(cdef, aconf.name) if 'D' in aconf.actions else '',
        _data_id_request(cdef, aconf.name) if 'R' in aconf.actions else '',
        _data_upsert_request(cdef, aconf.name) if all(element in aconf.actions for element in ['C','U']) else '',
        _data_create_many_request(cdef, aconf.name) if 'C' in aconf.actions else '',
        _data_update_many_request(cdef, aconf.name) if 'U' in aconf.actions else '',
        _data_delete_many_request(cdef, aconf.name) if 'D' in aconf.actions else '',
        _data_find_request(cdef, aconf.name) if 'L' in aconf.actions else '',
        _data_sign_in_request(cdef, aconf.name) if class_needs_session(cdef) else '',
        _data_client(cdef, aconf)
    ], 2)


def _data_find_request_nums(cdef: CDef, method_name: str) -> str:
    return f"""
    public func {method_name}(_ {method_name}: Int) -> {to_list_request(cdef)} {'{'}
        if query == nil {'{'} query = {to_list_query(cdef)}() {'}'}
        query = query!.{method_name}({method_name})
        return self
    {'}'}

    public func {method_name}(_ {method_name}: Int) async throws -> [{to_result(cdef)}] {'{'}
        return try await self.{method_name}({method_name}).exec()
    {'}'}""".strip('\n')


def _data_find_request_method(cdef: CDef) -> str:
    return f"""
    public func order(_ order: {to_sort_orders(cdef)}) -> {to_list_request(cdef)} {'{'}
        if query == nil {'{'} query = {to_list_query(cdef)}() {'}'}
        query = query!.order(order)
        return self
    {'}'}

    public func order(_ orders: [{to_sort_orders(cdef)}]) -> {to_list_request(cdef)} {'{'}
        if query == nil {'{'} query = {to_list_query(cdef)}() {'}'}
        query = query!.order(orders)
        return self
    {'}'}

    public func order(_ order: {to_sort_orders(cdef)}) async throws -> [{to_result(cdef)}] {'{'}
        return try await self.order(order).exec()
    {'}'}

    public func order(_ orders: [{to_sort_orders(cdef)}]) async throws -> [{to_result(cdef)}] {'{'}
        return try await self.order(orders).exec()
    {'}'}

{join_lines(map(lambda n: _data_find_request_nums(cdef, n), ['skip', 'limit', 'pageSize', 'pageNo']), 2)}
""".strip('\n')


def _data_query_request_common(
    cdef: CDef,
    single: bool = True,
    is_create_many: bool = False,
    is_sign_in: bool = False) -> str:
    return join_lines([f"""
    public func pick(_ picks: [{to_result_picks(cdef)}]) -> Self {'{'}
        if query == nil {'{'} query = {to_single_query(cdef) if single or is_create_many else to_list_query(cdef)}() {'}'}
        query = query!.pick(picks)
        return self
    {'}'}

    public func pick(_ picks: [{to_result_picks(cdef)}]) async throws -> {to_session(cdef, 'swift') if is_sign_in else to_result(cdef) if single else to_list_result(cdef)} {'{'}
        return try await self.pick(picks).exec()
    {'}'}

    public func omit(_ omits: [{to_result_picks(cdef)}]) -> Self {'{'}
        if query == nil {'{'} query = {to_single_query(cdef) if single or is_create_many else to_list_query(cdef)}() {'}'}
        query = query!.omit(omits)
        return self
    {'}'}

    public func omit(_ omits: [{to_result_picks(cdef)}]) async throws -> {to_session(cdef, 'swift') if is_sign_in else to_result(cdef) if single else to_list_result(cdef)} {'{'}
        return try await self.omit(omits).exec()
    {'}'}""".strip('\n'), _data_query_request_includes(cdef, single, is_create_many, is_sign_in)], 2)


def _data_query_request_include(cdef: CDef, item: tuple[str, str], single: bool = True, is_create_many: bool = False, is_sign_in: bool = False) -> str:
    return f"""
    public func include(_ ref: {to_include_key(cdef.name, item[0])}, _ query: {item[1]}? = nil) -> Self {'{'}
        if self.query == nil {'{'} self.query = {to_single_query(cdef) if single or is_create_many else to_list_query(cdef)}() {'}'}
        self.query = self.query!.include(ref, query)
        return self
    {'}'}

    public func include(_ ref: {to_include_key(cdef.name, item[0])}, _ query: {item[1]}? = nil) async throws -> {to_session(cdef, 'swift') if is_sign_in else to_result(cdef) if single else to_list_result(cdef)} {'{'}
        if self.query == nil {'{'} self.query = {to_single_query(cdef) if single or is_create_many else to_list_query(cdef)}() {'}'}
        self.query = self.query!.include(ref, query)
        return try await self.exec()
    {'}'}
    """.strip('\n')


def _data_query_request_includes(cdef: CDef, single: bool = True, is_create_many: bool = False, is_sign_in: bool = False) -> str:
    items = class_include_items(cdef)
    if len(items) == 0:
        return ''
    return join_lines(map(lambda i: _data_query_request_include(cdef, i, single, is_create_many, is_sign_in), items), 2)


def _data_create_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_create_request(cdef)} {'{'}",
        f"    internal var input: {to_create_input(cdef)}",
        f"    internal var query: {to_single_query(cdef)}?",
        '\n',
        f"    internal init(input: {to_create_input(cdef)}, query: {to_single_query(cdef)}? = nil) {'{'}",
        '        self.input = input',
        '        self.query = query'
        '    }',
        '\n',
        f"    internal func exec() async throws -> {to_result(cdef)} {'{'}",
        f"        return try await RequestManager.shared.post(",
        f'            url: "/{name}", input: input, query: query',
        f"        )",
        "    }",
        '\n',
        _data_query_request_common(cdef),
        '}'
    ], 1)


def _data_update_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_update_request(cdef)} {'{'}",
        "    internal var id: String",
        f"    internal var input: {to_update_input(cdef)}",
        f"    internal var query: {to_single_query(cdef)}?",
        '\n',
        f"    internal init(id: String, input: {to_update_input(cdef)}, query: {to_single_query(cdef)}? = nil) {'{'}",
        '        self.id = id',
        '        self.input = input',
        '        self.query = query'
        '    }',
        '\n',
        f"    internal func exec() async throws -> {to_result(cdef)} {'{'}",
        f"        return try await RequestManager.shared.patch(",
        f'            url: "/{name}/\(id)", input: input, query: query',
        f"        )",
        "    }",
        '\n',
        _data_query_request_common(cdef),
        '}'
    ], 1)


def _data_delete_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_delete_request(cdef)} {'{'}",
        "    internal var id: String",
        "\n",
        "    internal init(id: String) {",
        '        self.id = id',
        '    }',
        "\n",
        "    internal func exec() async throws {",
        "        return try await RequestManager.shared.delete(",
        f'            url: "/{name}/\(id)"',
        "        )",
        "    }",
        "}"
    ], 1)


def _data_id_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_id_request(cdef)} {'{'}",
        "    internal var id: String",
        f"    internal var query: {to_single_query(cdef)}?",
        '\n',
        f"    internal init(id: String, query: {to_single_query(cdef)}? = nil) {'{'}",
        '        self.id = id',
        '        self.query = query',
        '    }',
        '\n',
        f"    internal func exec() async throws -> {to_result(cdef)} {'{'}",
        f"        return try await RequestManager.shared.get(",
        f'            url: "/{name}/\(id)", query: query',
        f"        )!",
        "    }",
        '\n',
        _data_query_request_common(cdef),
        '}'
    ], 1)


def _data_upsert_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_upsert_request(cdef)} {'{'}",
        f"    internal var input: {to_query_data(cdef)}",
        f"    internal var query: {to_seek_query(cdef)}?",
        '\n',
        f"    internal init(input: {to_query_data(cdef)}, query: {to_seek_query(cdef)}? = nil) {'{'}",
        '        self.input = input',
        '        self.query = query',
        '    }',
        '\n',
        f"    internal func exec() async throws -> {to_result(cdef)} {'{'}",
        f"        return try await RequestManager.shared.post(",
        f'            url: "/{name}",',
        f'            input: {to_many_request_type(cdef)}.upsert.getContent(input: self.input),',
        "            query: self.query",
        "        )",
        "    }",
        '}'
    ], 1)


def _data_create_many_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_create_many_request(cdef)} {'{'}",
        f"    internal var input: [{to_create_input(cdef)}]",
        f"    internal var query: {to_single_query(cdef)}?",
        '\n',
        f"    internal init(input: [{to_create_input(cdef)}], query: {to_single_query(cdef)}? = nil) {'{'}",
        '        self.input = input',
        '        self.query = query'
        '    }',
        '\n',
        f"    internal func exec() async throws -> [{to_result(cdef)}] {'{'}",
        f"        return try await RequestManager.shared.post(",
        f'            url: "/{name}", input: {to_many_request_type(cdef)}.create.getContent(input: self.input), query: query',
        f"        )",
        "    }",
        '\n',
        _data_query_request_common(cdef, False, True),
        '}'
    ], 1)


def _data_update_many_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_update_many_request(cdef)} {'{'}",
        f"    internal var input: {to_query_data(cdef)}",
        f"    internal var query: {to_seek_query(cdef)}?",
        '\n',
        f"    internal init(input: {to_query_data(cdef)}, query: {to_seek_query(cdef)}? = nil) {'{'}",
        '        self.input = input',
        '        self.query = query',
        '    }',
        '\n',
        f"    internal func exec() async throws -> [{to_result(cdef)}] {'{'}",
        f"        return try await RequestManager.shared.patch(",
        f'            url: "/{name}", input: {to_many_request_type(cdef)}.update.getContent(input: self.input), query: self.query',
        f"        )",
        "    }",
        '}'
    ], 1)


def _data_delete_many_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_delete_many_request(cdef)} {'{'}",
        f"    internal var query: {to_seek_query(cdef)}?",
        "\n",
        f"    internal init(query: {to_seek_query(cdef)}? = nil) {'{'}",
        '        self.query = query',
        '    }',
        "\n",
        "    internal func exec() async throws {",
        "        return try await RequestManager.shared.delete(",
        f'            url: "/{name}",',
        '            query: self.query',
        "        )",
        "    }",
        "}"
    ], 1)


def _data_find_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_list_request(cdef)} {'{'}",
        f"    internal var query: {to_list_query(cdef)}?",
        '\n',
        f"    internal init(query: {to_list_query(cdef)}? = nil) {'{'}",
        '        self.query = query',
        '    }',
        '\n',
        f"    internal func exec() async throws -> [{to_result(cdef)}] {'{'}",
        f"        return try await RequestManager.shared.get(",
        f'            url: "/{name}", query: query',
        f"        )!",
        "    }",
        '\n',
        _data_find_request_method(cdef),
        _data_query_request_common(cdef, False),
        '}'
    ], 1)


def _data_sign_in_request(cdef: CDef, name: str) -> str:
    return join_lines([
        f"public class {to_sign_in_request(cdef)} {'{'}",
        f"    internal var input: {to_session_input(cdef)}",
        f"    internal var query: {to_single_query(cdef)}?",
        '\n',
        f"    internal init(input: {to_session_input(cdef)}, query: {to_single_query(cdef)}? = nil) {'{'}",
        '        self.input = input',
        '        self.query = query',
        '    }',
        '\n',
        f"    internal func exec() async throws -> {to_session(cdef, 'swift')} {'{'}",
        f"        SessionManager.shared.session = try await RequestManager.shared.post(",
        f'            url: "/{name}/session", input: input, query: query',
        "        )",
        "        return SessionManager.shared.session!",
        "    }",
        '\n',
        _data_query_request_common(cdef, is_sign_in=True),
        '}'
    ], 1)


def _data_client(cdef: CDef, aconf: AConf) -> str:
    return join_lines([
        f'public struct {to_client(cdef)} {"{"}',
        '\n',
        '    fileprivate init() { }',
        '\n',
        join_lines([
            _data_client_creates(cdef, aconf),
            _data_client_updates(cdef, aconf),
            _data_client_delete(cdef, aconf),
            _data_client_ids(cdef, aconf),
            _data_client_finds(cdef, aconf),
            _data_client_upsert(cdef, aconf),
            _data_client_create_many(cdef, aconf),
            _data_client_update_many(cdef, aconf),
            _data_client_delete_many(cdef, aconf),
            _data_client_sign_in(cdef),
        ], 2),
        '}'
    ], 1)


def _data_client_create_2(cdef: CDef, items: list[CodableClassItem]) -> str:
    if len(items) == 0:
        return join_lines([
            f'    public func create() -> {to_create_request(cdef)} {"{"}',
            f'        let input = {to_create_input(cdef)}()',
            '        return create(input)',
            '    }'
        ], 1)
    last = len(items) - 1
    return join_lines([
        f'    public func create(',
        *map(lambda i: f"        {i[1][2]}: {i[1][3]}{'? = nil' if i[1][4] else ''}{'' if i[0] == last else ', '}", enumerate(items)),
        f'    ) -> {to_create_request(cdef)} {"{"}',
        f'        let input = {to_create_input(cdef)}(',
        *map(lambda i: f"            {i[1][2]}: {i[1][2]}{'' if i[0] == last else ','}", enumerate(items)),
        '        )',
        '        return create(input)',
        '    }'
    ], 1)


def _data_client_create_4(cdef: CDef, items: list[CodableClassItem]) -> str:
    if len(items) == 0:
        return join_lines([
            f'    public func create() async throws -> {to_result(cdef)} {"{"}',
            f'        let request: {to_create_request(cdef)} = self.create()',
            '        return try await request.exec()',
            '    }'
        ], 1)
    last = len(items) - 1
    return join_lines([
        f'    public func create(',
        *map(lambda i: f"        {i[1][2]}: {i[1][3]}{'? = nil' if i[1][4] else ''}{'' if i[0] == last else ', '}", enumerate(items)),
        f'    ) async throws -> {to_result(cdef)} {"{"}',
        f'        let request: {to_create_request(cdef)} = self.create(',
        *map(lambda i: f"            {i[1][2]}: {i[1][2]}{'' if i[0] == last else ','}", enumerate(items)),
        '        )',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_creates(cdef: CDef, aconf: AConf) -> str:
    if 'C' not in aconf.actions:
        return ''
    input_items = class_create_input_items(cdef)
    return join_lines([
        f'    public func create(_ input: {to_create_input(cdef)}) -> {to_create_request(cdef)} {"{"}',
        f'        return {to_create_request(cdef)}(input: input)',
        '    }',
        '\n',
        _data_client_create_2(cdef, input_items),
        '\n',
        f'    public func create(_ input: {to_create_input(cdef)}) async throws -> {to_result(cdef)} {"{"}',
        f'        let request: {to_create_request(cdef)} = self.create(input)',
        '        return try await request.exec()',
        '    }',
        '\n',
        _data_client_create_4(cdef, input_items)
    ], 1)


def _data_client_update_2(cdef: CDef, items: list[CodableClassItem]) -> str:
    if len(items) == 0:
        return join_lines([
            f'    public func update(_ id: String) -> {to_update_request(cdef)} {"{"}',
            f'        let input = {to_update_input(cdef)}()',
            '        return update(id, input)',
            '    }'
        ], 1)
    last = len(items) - 1
    return join_lines([
        f'    public func update(',
        '        _ id: String,',
        *map(lambda i: f"        {i[1][2]}: {i[1][3]}{'? = nil' if i[1][4] else ''}{'' if i[0] == last else ', '}", enumerate(items)),
        f'    ) -> {to_update_request(cdef)} {"{"}',
        f'        let input = {to_update_input(cdef)}(',
        *map(lambda i: f"            {i[1][2]}: {i[1][2]}{'' if i[0] == last else ','}", enumerate(items)),
        '        )',
        '        return update(id, input)',
        '    }'
    ], 1)


def _data_client_update_4(cdef: CDef, items: list[CodableClassItem]) -> str:
    if len(items) == 0:
        return join_lines([
            f'    public func update(_ id: String) async throws -> {to_result(cdef)} {"{"}',
            f'        let request: {to_update_request(cdef)} = self.update(id)',
            '        return try await request.exec()',
            '    }'
        ], 1)
    last = len(items) - 1
    return join_lines([
        f'    public func update(',
        '        _ id: String,',
        *map(lambda i: f"        {i[1][2]}: {i[1][3]}{'? = nil' if i[1][4] else ''}{'' if i[0] == last else ', '}", enumerate(items)),
        f'    ) async throws -> {to_result(cdef)} {"{"}',
        f'        let request: {to_update_request(cdef)} = self.update(',
        '            id,',
        *map(lambda i: f"            {i[1][2]}: {i[1][2]}{'' if i[0] == last else ','}", enumerate(items)),
        '        )',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_updates(cdef: CDef, aconf: AConf) -> str:
    if 'U' not in aconf.actions:
        return ''
    input_items = class_update_input_items(cdef)
    return join_lines([
        f'    public func update(_ id: String, _ input: {to_update_input(cdef)}) -> {to_update_request(cdef)} {"{"}',
        f'        return {to_update_request(cdef)}(id: id, input: input)',
        '    }',
        '\n',
        _data_client_update_2(cdef, input_items),
        '\n',
        f'    public func update(_ id: String, _ input: {to_update_input(cdef)}) async throws -> {to_result(cdef)} {"{"}',
        f'        let request: {to_update_request(cdef)} = self.update(id, input)',
        '        return try await request.exec()',
        '    }',
        '\n',
        _data_client_update_4(cdef, input_items)
    ], 1)


def _data_client_delete(cdef: CDef, aconf: AConf) -> str:
    if 'D' not in aconf.actions:
        return ''
    return join_lines([
        '    public func delete(_ id: String) async throws {',
        f'        let request = {to_delete_request(cdef)}(id: id)',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_ids(cdef: CDef, aconf: AConf) -> str:
    if 'R' not in aconf.actions:
        return ''
    return join_lines([
        f'    public func id(_ id: String) -> {to_id_request(cdef)} {"{"}',
        f'        return {to_id_request(cdef)}(id: id)',
        '    }',
        f'    public func id(_ id: String) async throws -> {to_result(cdef)} {"{"}',
        f'        let request = {to_id_request(cdef)}(id: id)',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_find_2(cdef: CDef, items: list[tuple[str, str]]) -> str:
    last = len(items) - 1
    return join_lines([
        '    public func find(',
        *map(lambda i: f"        {i[1][0]}: {i[1][1]}? = nil{'' if i[0] == last else ','}", enumerate(items)),
        f'    ) -> {to_list_request(cdef)} {"{"}',
        f'        let query = {to_list_query(cdef)}()',
        *map(lambda i: f"        query.{i[0]} = {i[0]}", items),
        f'        return {to_list_request(cdef)}(query: query)',
        '    }'
    ], 1)


def _data_client_find_4(cdef: CDef, items: list[tuple[str, str]]) -> str:
    last = len(items) - 1
    return join_lines([
        '    public func find(',
        *map(lambda i: f"        {i[1][0]}: {i[1][1]}? = nil{'' if i[0] == last else ','}", enumerate(items)),
        f'    ) async throws -> {to_list_result(cdef)} {"{"}',
        f'        let query = {to_list_query(cdef)}()',
        *map(lambda i: f"        query.{i[0]} = {i[0]}", items),
        f'        let request = {to_list_request(cdef)}(query: query)',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_finds(cdef: CDef, aconf: AConf) -> str:
    if 'L' not in aconf.actions:
        return ''
    query_items = list_query_items(cdef)
    return join_lines([
        f'    public func find(_ query: {to_list_query(cdef)}? = nil) -> {to_list_request(cdef)} {"{"}',
        f'        return {to_list_request(cdef)}(query: query)',
        '    }',
        '\n',
        _data_client_find_2(cdef, query_items),
        '\n',
        f'    public func find(_ query: {to_list_query(cdef)}? = nil) async throws -> {to_list_result(cdef)} {"{"}',
        f'        let request = {to_list_request(cdef)}(query: query)',
        '        return try await request.exec()',
        '    }',
        '\n',
        _data_client_find_4(cdef, query_items)
    ], 1)


def _data_client_upsert(cdef: CDef, aconf: AConf) -> str:
    if not all( element in aconf.actions for element in ['C','U']):
        return ''
    return join_lines([
        f'    public func upsert(query: {to_seek_query(cdef)}, data: {to_update_input(cdef)}) async throws -> {to_result(cdef)} {"{"}',
        f'        let input = {to_query_data(cdef)}(_query: query, _data: data)',
        f'        let request = {to_upsert_request(cdef)}(input: input)',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_create_many(cdef: CDef, aconf: AConf) -> str:
    if 'C' not in aconf.actions:
        return ''
    return join_lines([
        f'    public func createMany(input: [{to_create_input(cdef)}], query: {to_single_query(cdef)}? = nil) -> {to_create_many_request(cdef)} {"{"}',
        f'        return {to_create_many_request(cdef)}(input: input, query: query)',
        '    }',
        '\n',
        f'    public func createMany(input: [{to_create_input(cdef)}], query: {to_single_query(cdef)}? = nil) async throws -> [{to_result(cdef)}] {"{"}',
        f'        let request = {to_create_many_request(cdef)}(input: input, query: query)',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_update_many(cdef: CDef, aconf: AConf) -> str:
    if 'U' not in aconf.actions:
        return ''
    return join_lines([
        f'    public func updateMany(query: {to_seek_query(cdef)}, data: {to_update_input(cdef)}) async throws -> [{to_result(cdef)}] {"{"}',
        f'        let input = {to_query_data(cdef)}(_query: query, _data: data)',
        f'        let request = {to_update_many_request(cdef)}(input: input)',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_delete_many(cdef: CDef, aconf: AConf) -> str:
    if 'D' not in aconf.actions:
        return ''
    return join_lines([
        f'    public func delete(_ query: {to_seek_query(cdef)}? = nil) async throws {"{"}',
        f'        let request = {to_delete_many_request(cdef)}(query: query)',
        '        return try await request.exec()',
        '    }'
    ], 1)


def _data_client_sign_in(cdef: CDef) -> str:
    if not class_needs_session(cdef):
        return ''
    return join_lines([
        f'    public func signIn(input: {to_session_input(cdef)}, query: {to_single_query(cdef)}? = nil) async throws -> {to_session(cdef, "swift")} {"{"}',
        f'        let request = {to_sign_in_request(cdef)}(input: input, query: query)',
        '        return try await request.exec()',
        '    }'
    ], 1)



def gitignore_content() -> str:
    return """
# packages
/packages

# python
__pycache__

# venv
venv
.venv

# dotenv
.env

# mypy
.mypy_cache

# vscode
.vscode

# idea
.idea
*.iml
    """.strip() + '\n'



def date_query():
    return """
public enum DateQuery: Codable {
    case eq(_ value: Date)
    case neq(_ value: Date)
    case null(_ value: Bool)
    case gt(_ value: Date)
    case gte(_ value: Date)
    case lt(_ value: Date)
    case lte(_ value: Date)
    case after(_ value: Date)
    case before(_ value: Date)
    case or(_ values: [DateQuery])
    case and(_ values: [DateQuery])

    public enum CodingKeys: String, CodingKey {
        case eq = "_eq"
        case neq = "_neq"
        case null = "_null"
        case gt = "_gt"
        case gte = "_gte"
        case lt = "_lt"
        case lte = "_lte"
        case after = "_after"
        case before = "_before"
        case or = "_or"
        case and = "_and"
    }

    public init(from decoder: Decoder) throws {
        let container = try! decoder.container(keyedBy: CodingKeys.self)
        if container.contains(.eq) {
            self = .eq(try! container.decode(Date.self, forKey: .eq))
        } else if container.contains(.neq) {
            self = .neq(try! container.decode(Date.self, forKey: .neq))
        } else if container.contains(.null) {
            self = .null(try! container.decode(Bool.self, forKey: .null))
        } else if container.contains(.gt) {
            self = .gt(try! container.decode(Date.self, forKey: .gt))
        } else if container.contains(.gte) {
            self = .gte(try! container.decode(Date.self, forKey: .gte))
        } else if container.contains(.lt) {
            self = .lt(try! container.decode(Date.self, forKey: .lt))
        } else if container.contains(.lte) {
            self = .lte(try! container.decode(Date.self, forKey: .lte))
        } else if container.contains(.after) {
            self = .after(try! container.decode(Date.self, forKey: .after))
        } else if container.contains(.before) {
            self = .before(try! container.decode(Date.self, forKey: .before))
        } else if container.contains(.or) {
            self = .or(try! container.decode([DateQuery].self, forKey: .or))
        } else if container.contains(.and) {
            self = .and(try! container.decode([DateQuery].self, forKey: .and))
        } else {
            self = .eq(Date())
        }
    }

    public func encode(to encoder: Encoder) throws {
            var container = encoder.container(keyedBy: CodingKeys.self)
            switch self {
            case .eq(let value):
                try! container.encode(value, forKey: .eq)
            case .neq(let value):
                try! container.encode(value, forKey: .neq)
            case .null(let value):
                try! container.encode(value, forKey: .null)
            case .gt(let value):
                try! container.encode(value, forKey: .gt)
            case .gte(let value):
                try! container.encode(value, forKey: .gte)
            case .lt(let value):
                try! container.encode(value, forKey: .lt)
            case .lte(let value):
                try! container.encode(value, forKey: .lte)
            case .after(let value):
                try! container.encode(value, forKey: .after)
            case .before(let value):
                try! container.encode(value, forKey: .before)
            case .or(let value):
                try! container.encode(value, forKey: .or)
            case .and(let value):
                try! container.encode(value, forKey: .and)
            }
        }
}
    """.strip() + "\n"



from __future__ import annotations
import re
from os import getcwd
from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from jsonclasses_cli.new import new


class TestNew(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.temp_dir = TemporaryDirectory()
        cls.data_path = Path(getcwd()) / 'tests' / 'data_new'

    @classmethod
    def tearDownClass(cls) -> None:
        cls.temp_dir.cleanup()

    def setUp(self) -> None:
        self.temp_path = Path(str(self.temp_dir.name)) / "new_app"


    def test_new_create_all_filed(self) -> None:
        new(self.temp_path, None, None, None, None, None, True)
        app = self.temp_path / 'app.py'
        requirements = self.temp_path / 'requirements.txt'
        config = self.temp_path / 'config.json'
        mypy = self.temp_path / 'mypy.ini'
        gitignore = self.temp_path / '.gitignore'
        read_me = self.temp_path / 'README.md'
        self.assertTrue(app.is_file(), app.name)
        self.assertTrue(requirements.is_file(), requirements.name)
        self.assertTrue(config.is_file(), config.name)
        self.assertTrue(mypy.is_file(), mypy.name)
        self.assertTrue(gitignore.is_file(), gitignore.name)
        self.assertTrue(read_me.is_file(), read_me.name)

    def test_new_app_content_without_admin_and_user(self) -> None:
        new(self.temp_path, None, None, None, None, None, True)
        result = self.temp_path / 'app.py'
        expect = self.data_path / 'app.py'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_new_app_content_with_user(self) -> None:
        new(self.temp_path, None, True, None, None, None, True)
        result = self.temp_path / 'app.py'
        expect = self.data_path / 'app_with_u.py'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_new_app_content_with_admin(self) -> None:
        new(self.temp_path, None, None, True, None, None, True)
        result = self.temp_path / 'app.py'
        expect = self.data_path / 'app_with_a.py'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_new_app_content_with_user_and_admin(self) -> None:
        new(self.temp_path, None, True, True, None, None, True)
        result = self.temp_path / 'app.py'
        expect = self.data_path / 'app_with_ua.py'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_new_genarate_content_of_config(self) -> None:
        new(self.temp_path, None, True, True, None, None, True)
        result = (self.temp_path / 'config.json').read_text()
        expect = (self.data_path / 'config.json').read_text()
        filter_result = re.sub(r'\"secretKey\": \".*\"', '', result)
        self.assertEqual(filter_result, expect)

    def test_new_genarate_content_of_requirements(self) -> None:
        new(self.temp_path, None, False, False, None, None, True)
        result = (self.temp_path / 'requirements.txt').read_text()
        expect = (self.data_path / 'requirements.txt').read_text()
        self.assertEqual(result, expect)

    def test_new_genarate_content_of_requirements_with_user_or_admin(self) -> None:
        new(self.temp_path, None, True, False, None, None, True)
        result = (self.temp_path / 'requirements.txt').read_text()
        expect = (self.data_path / 'requirements_with_ua.txt').read_text()
        self.assertEqual(result, expect)

    def test_new_genarate_content_of_mypy_gitignore_and_readme(self) -> None:
        new(self.temp_path, None, None, None, None, None, True)
        mypy = self.temp_path / 'mypy.ini'
        gitignore = self.temp_path / '.gitignore'
        read_me = self.temp_path / 'README.md'
        expect_mypy = self.data_path / 'mypy.ini'
        expect_gitignore = self.data_path / '.gitignore'
        expect_read_me = self.data_path / 'README.md'
        self.assertEqual(mypy.read_text(), expect_mypy.read_text())
        self.assertEqual(gitignore.read_text(), expect_gitignore.read_text())
        self.assertEqual(read_me.read_text(), expect_read_me.read_text())



from typing import Literal
from jsonclasses.fdef import FType, FDef
from ...utils.package_utils import is_field_link


def jtype_to_ts_type(fdef: FDef, mode: Literal['C', 'U', 'R', 'Q'], is_link: bool = False) -> str:
    match fdef.ftype:
        case FType.STR:
            if mode == 'Q':
                return 'StringQuery'
            else:
                return 'string'
        case FType.INT:
            if mode == 'Q':
                return 'NumberQuery'
            else:
                return 'number'
        case FType.FLOAT:
            if mode == 'Q':
                return 'NumberQuery'
            else:
                return 'number'
        case FType.BOOL:
            if mode == 'Q':
                return 'BooleanQuery'
            else:
                return 'boolean'
        case FType.DATE:
            if mode == 'Q':
                return 'DateQuery'
            elif mode == 'R':
                return 'string'
            else:
                return 'Date'
        case FType.DATETIME:
            if mode == 'Q':
                return 'DateQuery'
            elif mode == 'R':
                return 'string'
            else:
                return 'Date'
        case FType.ENUM:
            return fdef.enum_class.__name__
        case FType.LIST:
            return jtype_to_ts_type(fdef.item_types.fdef, mode, is_field_link(fdef)) + '[]'
        case FType.DICT:
            return '{[key: string]: ' + jtype_to_ts_type(fdef.item_types.fdef, mode) + '}'
        case FType.INSTANCE:
            if mode == 'R':
                return fdef.inst_cls.__name__
            elif mode == 'C':
                result = fdef.inst_cls.__name__ + 'CreateInput'
                if is_link:
                    return f'({result} | Link)'
                return result
            elif mode == 'U':
                result = fdef.inst_cls.__name__ + 'UpdateInput'
                if is_link:
                    return f'({result} | Link | UnLink)'
                return result
            else:
                return 'never'
        case FType.ANY:
            return 'any'
        case FType.UNION:
            return " | ".join([jtype_to_ts_type(t.fdef, mode) for t in fdef.raw_union_types])
        case None:
            return "never"



def session_manager() -> str:
    return """
public struct SessionManager {

    public static var shared = SessionManager()

    @UserDefault(key: "session") public fileprivate(set) var session: Session?

    private init() { }
}
    """.strip() + '\n'



from os import getcwd
from pathlib import Path
from rich import print

def write_file(path: Path, content: str, silent: bool = False) -> None:
    if not path.parent.exists():
        path.parent.mkdir(parents=True)
    with open(path, 'w') as file:
        if not silent:
            print(f"[bold green]CREATE[/bold green] {path.relative_to(getcwd())}")
        file.write(content)



from pathlib import Path
from jsonclasses.cgraph import CGraph
from .main_program_content import main_program_content
from .gitignore_content import gitignore_content
from .package_content import package_content
from .readme_content import readme_content
from ...utils.write_file import write_file


def swift(dest: Path, cgraph: CGraph, silent: bool = False):
    dest = _create_dest_dir_if_needed(dest)
    _generate_main_program_file(dest, cgraph, silent)
    _generate_package_file(dest, silent)
    _generate_readme_file(dest, silent)
    _generate_gitignore_file(dest, silent)


def _create_dest_dir_if_needed(dest: Path) -> Path:
    dest = dest / 'packages' / 'swift'
    if not dest.is_dir():
        dest.mkdir(parents=True)
    return dest


def _generate_main_program_file(dest: Path, cgraph: CGraph, silent: bool = False):
    write_file(dest / 'Sources' / 'API' / 'API.swift', main_program_content(cgraph), silent)


def _generate_package_file(dest: Path, silent: bool = False):
    write_file(dest / 'Package.swift', package_content(), silent)


def _generate_gitignore_file(dest: Path, silent: bool = False):
    write_file(dest / '.gitignore', gitignore_content(), silent)


def _generate_readme_file(dest: Path, silent: bool = False):
    write_file(dest / 'README.md', readme_content(), silent)



from pathlib import Path
from rich import print
from rich.prompt import Prompt
from .app_content import app_content
from .req_content import req_content
from .conf_content import conf_content
from .mypy_content import mypy_content
from .gitignore_content import gitignore_content
from .readme_content import readme_content
from ..utils.yesno import yesno
from ..utils.write_file import write_file
from ..utils.run import run


def new(dest: Path,
        interactive: bool | None,
        include_user: bool | None,
        include_admin: bool | None,
        git_init: bool | None,
        venv: bool | None,
        silent: bool = False):
    if interactive:
        if include_user is None:
            include_user = yesno(Prompt.ask('Do you want a user model?', choices=['Yes', 'No'], default='Yes'))
        if include_admin is None:
            include_admin = yesno(Prompt.ask('Do you want an admin model?', choices=['Yes', 'No'], default='Yes'))
        if git_init is None:
            git_init = yesno(Prompt.ask('Init git repo?', choices=['Yes', 'No'], default='Yes'))
        if venv is None:
            venv = yesno(Prompt.ask('Create a virtual env?', choices=['Yes', 'No'], default='Yes'))
    else:
        if include_user is None:
            include_user = False
        if include_admin is None:
            include_admin = False
        if git_init is None:
            git_init = True
        if venv is None:
            venv = True
    write_file(dest / 'app.py', app_content(include_user=include_user, include_admin=include_admin), silent)
    write_file(dest / 'requirements.txt', req_content(include_user=include_user, include_admin=include_admin), silent)
    write_file(dest / 'config.json', conf_content(dest.name), silent)
    write_file(dest / 'mypy.ini', mypy_content(), silent)
    write_file(dest / '.gitignore', gitignore_content(), silent)
    write_file(dest / 'README.md', readme_content(dest), silent)
    if git_init:
        if not (dest / '.git').is_dir():
            run('git init', silent)
    if venv:
        if not ((dest / 'venv').is_dir() or (dest / '.venv').is_dir()):
            dest_venv = dest / '.venv'
            run(f'python3 -m venv {dest_venv}', silent)
    if venv:
        venv_dir = dest / 'venv' if (dest / 'venv').is_dir() else dest / '.venv'
        venv_act = venv_dir / 'bin/activate'
        run(f'source {venv_act}; pip install -r requirements.txt', silent)
    else:
        run('pip install -r requirements.txt', silent)
    if not silent:
        print("🎉[green]Project is successfully created.[/green]")
        print("\n    Run 'uvicorn app:app --reload' to start the development server.\n")



from typing import Literal
from jsonclasses.fdef import FDef, FType
from ...utils.package_utils import is_field_link


def jtype_to_swift_type(fdef: FDef, mode: Literal['C', 'U', 'R', 'Q'], is_link: bool = False) -> str:
    match fdef.ftype:
        case FType.STR:
            if mode == 'Q':
                return 'StringQuery'
            else:
                return 'String'
        case FType.INT:
            if mode == 'Q':
                return 'IntQuery'
            else:
                return 'Int'
        case FType.FLOAT:
            if mode == 'Q':
                return 'FloatQuery'
            else:
                return 'Float'
        case FType.BOOL:
            if mode == 'Q':
                return 'BoolQuery'
            else:
                return 'Bool'
        case FType.DATE:
            if mode == 'Q':
                return 'DateQuery'
            else:
                return 'Date'
        case FType.DATETIME:
            if mode == 'Q':
                return 'DateQuery'
            else:
                return 'Date'
        case FType.ENUM:
            return fdef.enum_class.__name__
        case FType.LIST:
            return '[' + jtype_to_swift_type(fdef.item_types.fdef, mode, is_field_link(fdef)) + ']'
        case FType.DICT:
            return '[String: ' + jtype_to_swift_type(fdef.item_types.fdef, mode) + ']'
        case FType.INSTANCE:
            if mode == 'R':
                return fdef.inst_cls.__name__
            elif mode == 'C':
                result = fdef.inst_cls.__name__ + 'CreateInput'
                if is_link:
                    return f'CreateOrLink<{result}>'
                return result
            elif mode == 'U':
                result = fdef.inst_cls.__name__ + 'UpdateInput'
                if is_link:
                    return f'UpdateOrLink<{result}>'
                return result
            else:
                return 'never'
        case FType.ANY:
            return 'Never'
        case FType.UNION:
            return "Never"
        case None:
            return "Never"



from enum import Enum
from inflection import camelize
from .codable_enum import codable_enum, codable_enum_item


def data_enum(enum: type[Enum]) -> str:
    return codable_enum(enum.__name__, 'String', map(_data_enum_item, enum))


def _data_enum_item(option: Enum):
    name = camelize(option.name.lower())
    value = option.name
    return codable_enum_item(name, 'String', value)



from typing import cast
from jsonclasses.cdef import CDef
from jsonclasses_cli.utils.package_utils import to_client
from jsonclasses_server.aconf import AConf
from jsonclasses.cgraph import CGraph
from ...utils.join_lines import join_lines
from ...utils.package_utils import session_input_cdefs, to_session, to_session_input, to_sign_in_request, to_single_query
from inflection import camelize,pluralize

def class_api(cgraph: CGraph, use_session:bool) -> str:
    return join_lines([
        'class API {',
        *map(lambda c: _client_item(c), cgraph._map.values()),
        _session() if use_session else '',
        _sign_out() if use_session else '',
        '}'
    ], 2)


def _client_item(cdef: CDef) -> str:
    name  = camelize(pluralize(cdef.name))
    return join_lines([
        f"    get {name}(): {to_client(cdef)} {'{'}",
        f"        return new {to_client(cdef)}()",
        "    }"
    ])


def _session() -> str:
    return join_lines([
        '    get session(): SessionManager {',
        '       return SessionManager.share',
        '    }'
    ])


def _sign_out() -> str:
    return join_lines([
        '    signOut(): void {',
        '       SessionManager.share.clearSession()',
        '    }'
    ])



def gitignore_content() -> str:
    return """
lib

# Logs
logs
*.log
npm-debug.log*
yarn-debug.log*
yarn-error.log*
lerna-debug.log*

# Diagnostic reports (https://nodejs.org/api/report.html)
report.[0-9]*.[0-9]*.[0-9]*.[0-9]*.json

# Runtime data
pids
*.pid
*.seed
*.pid.lock

# Directory for instrumented libs generated by jscoverage/JSCover
lib-cov

# Coverage directory used by tools like istanbul
coverage
*.lcov

# nyc test coverage
.nyc_output

# Grunt intermediate storage (https://gruntjs.com/creating-plugins#storing-task-files)
.grunt

# Bower dependency directory (https://bower.io/)
bower_components

# node-waf configuration
.lock-wscript

# Compiled binary addons (https://nodejs.org/api/addons.html)
build/Release

# Dependency directories
node_modules/
jspm_packages/

# TypeScript v1 declaration files
typings/

# TypeScript cache
*.tsbuildinfo

# Optional npm cache directory
.npm

# Optional eslint cache
.eslintcache

# Microbundle cache
.rpt2_cache/
.rts2_cache_cjs/
.rts2_cache_es/
.rts2_cache_umd/

# Optional REPL history
.node_repl_history

# Output of 'npm pack'
*.tgz

# Yarn Integrity file
.yarn-integrity

# dotenv environment variables file
.env
.env.test

# parcel-bundler cache (https://parceljs.org/)
.cache

# Next.js build output
.next

# Nuxt.js build / generate output
.nuxt
dist

# Gatsby files
.cache/
# Comment in the public line in if your project uses Gatsby and *not* Next.js
# https://nextjs.org/blog/next-9-1#public-directory-support
# public

# vuepress build output
.vuepress/dist

# Serverless directories
.serverless/

# FuseBox cache
.fusebox/

# DynamoDB Local files
.dynamodb/

# TernJS port file
.tern-port
    """.strip() + '\n'



def string_query():
    return """
public enum StringQuery: Codable {
    case eq(_ value: String)
    case neq(_ value: String)
    case null(_ value: Bool)
    case gt(_ value: String)
    case gte(_ value: String)
    case lt(_ value: String)
    case lte(_ value: String)
    case contains(_ value: String, mode: Mode = .default)
    case prefix(_ value: String, mode: Mode = .default)
    case suffix(_ value: String, mode: Mode = .default)
    case match(_ value: String, mode: Mode = .default)
    case or(_ values: [StringQuery])
    case and(_ values: [StringQuery])

    public enum Mode: String, Codable {
        case `default` = "default"
        case caseInsensitive = "insensitive"
    }

    public enum CodingKeys: String, CodingKey {
        case eq = "_eq"
        case neq = "_neq"
        case null = "_null"
        case gt = "_gt"
        case gte = "_gte"
        case lt = "_lt"
        case lte = "_lte"
        case contains = "_contains"
        case prefix = "_prefix"
        case suffix = "_suffix"
        case match = "_match"
        case mode = "_mode"
        case or = "_or"
        case and = "_and"
    }

    public init(from decoder: Decoder) throws {
        let container = try! decoder.container(keyedBy: CodingKeys.self)
        if container.contains(.eq) {
            self = .eq(try! container.decode(String.self, forKey: .eq))
        } else if container.contains(.neq) {
            self = .neq(try! container.decode(String.self, forKey: .neq))
        }else if container.contains(.null) {
            self = .null(try! container.decode(Bool.self, forKey: .null))
        }  else if container.contains(.gt) {
            self = .gt(try! container.decode(String.self, forKey: .gt))
        } else if container.contains(.gte) {
            self = .gte(try! container.decode(String.self, forKey: .gte))
        } else if container.contains(.lt) {
            self = .lt(try! container.decode(String.self, forKey: .lt))
        } else if container.contains(.lte) {
            self = .lte(try! container.decode(String.self, forKey: .lte))
        } else if container.contains(.contains) {
            self = .contains(
                try! container.decode(String.self, forKey: .contains),
                mode: (try? container.decode(Mode.self, forKey: .mode)) ?? .default
            )
        } else if container.contains(.prefix) {
            self = .prefix(
                try! container.decode(String.self, forKey: .prefix),
                mode: (try? container.decode(Mode.self, forKey: .mode)) ?? .default
            )
        } else if container.contains(.suffix) {
            self = .suffix(
                try! container.decode(String.self, forKey: .suffix),
                mode: (try? container.decode(Mode.self, forKey: .mode)) ?? .default
            )
        } else if container.contains(.match) {
            self = .match(
                try! container.decode(String.self, forKey: .match),
                mode: (try? container.decode(Mode.self, forKey: .mode)) ?? .default
            )
        } else if container.contains(.or) {
            self = .or(try! container.decode([StringQuery].self, forKey: .or))
        } else if container.contains(.and) {
            self = .and(try! container.decode([StringQuery].self, forKey: .and))
        } else {
            self = .eq("")
        }
    }

    public func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        switch self {
        case .eq(let value):
            try! container.encode(value, forKey: .eq)
        case .neq(let value):
            try! container.encode(value, forKey: .neq)
        case .null(let value):
            try! container.encode(value, forKey: .null)
        case .gt(let value):
            try! container.encode(value, forKey: .gt)
        case .gte(let value):
            try! container.encode(value, forKey: .gte)
        case .lt(let value):
            try! container.encode(value, forKey: .lt)
        case .lte(let value):
            try! container.encode(value, forKey: .lte)
        case .contains(let value, let mode):
            try! container.encode(value, forKey: .contains)
            if mode != .default {
                try! container.encode(mode, forKey: .mode)
            }
        case .prefix(let value, let mode):
            try! container.encode(value, forKey: .prefix)
            if mode != .default {
                try! container.encode(mode, forKey: .mode)
            }
        case .suffix(let value, let mode):
            try! container.encode(value, forKey: .suffix)
            if mode != .default {
                try! container.encode(mode, forKey: .mode)
            }
        case .match(let value, let mode):
            try! container.encode(value, forKey: .match)
            if mode != .default {
                try! container.encode(mode, forKey: .mode)
            }
        case .or(let value):
            try! container.encode(value, forKey: .or)
        case .and(let value):
            try! container.encode(value, forKey: .and)
        }
    }
}
    """.strip() + "\n"



def request_manager(base_url: str,use_session:bool) -> str:
    return f"""
class RequestManager {"{"}

    static share = new RequestManager()

    #baseURL: string = "{base_url}"


    get headers() {"{"}
        const token = {'SessionManager.share.hasSession() ? SessionManager.share.getToken() : ' if use_session else ''} undefined 
        return token ? {"{"}
            headers: {"{"}
                'Authorization': `Bearer ${"{"}token{"}"}`
            {"}"}
        {"}"} : undefined
    {"}"}

    qs(val: any): string {"{"}
        if (!val) {"{"}
            return ''
        {"}"}
        if (Object.keys(val).length === 0) {"{"}
            return ''
        {"}"}
        return '?' + stringify(val)
    {"}"}

    async post<T, U, V>(url: string, input: T, query: V | undefined = undefined): Promise<U> {"{"}
        const response = await axios.post(this.#baseURL + url + this.qs(query), input, this.headers)
        return response.data.data
    {"}"}

    async patch<T, U, V>(url: string, input: T, query: V | undefined = undefined): Promise<U> {"{"}
        const response = await axios.patch(this.#baseURL + url + this.qs(query), input, this.headers)
        return response.data.data
    {"}"}

    async delete<V>(url: string, query: V | undefined = undefined): Promise<void> {"{"}
        await axios.delete(this.#baseURL + url + this.qs(query), this.headers)
        return
    {"}"}

    async get<U, V>(url: string, query: V | undefined = undefined): Promise<U> {"{"}
        const response = await axios.get(this.#baseURL + url + this.qs(query), this.headers)
        return response.data.data
    {"}"}
{"}"}
    """.strip() + '\n'



from pathlib import Path
from jsonclasses.cgraph import CGraph
from .package_json_content import package_json_content
from .tsconfig_json_content import tsconfig_json_content
from .main_program_content import main_program_content
from ...utils.write_file import write_file


def ts(dest: Path, cgraph: CGraph, silent: bool = False):
    package_dest = _create_dest_dir_if_needed(dest)
    _generate_main_program_file(package_dest, cgraph, silent)
    _generate_package_json_file(package_dest, dest, silent)
    _generate_tsconfig_json_file(package_dest, silent)
    _generate_gitignore_file(package_dest, silent)


def _create_dest_dir_if_needed(dest: Path) -> Path:
    dest = dest / 'packages' / 'ts'
    if not dest.is_dir():
        dest.mkdir(parents=True)
    return dest


def _generate_main_program_file(dest: Path, cgraph: CGraph, silent: bool = False):
    write_file(dest / 'src/index.ts', main_program_content(cgraph), silent)


def _generate_package_json_file(dest: Path, original_dest: Path, silent: bool = False):
    write_file(dest / 'package.json', package_json_content(original_dest), silent)


def _generate_tsconfig_json_file(dest: Path, silent: bool = False):
    write_file(dest / 'tsconfig.json', tsconfig_json_content(), silent)


def _generate_gitignore_file(dest: Path, silent: bool = False):
    write_file(dest / '.gitignore', tsconfig_json_content(), silent)



from random import sample
from jsonclasses_cli.utils.join_lines import join_lines


def conf_content(name: str) -> str:
    return join_lines([
        '{',
        _pymongo_conf(name),
        _operator_conf(),
        _package_conf(),
        '}'
    ])


def _operator_conf() -> str:
    return f"""
    "operator": {'{'}
        "secretKey": "{_randompuncs()}"
    {'}'},""".strip('\n')


def _randompuncs() -> str:
    ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
    ascii_uppercase = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    punctuation = r"""!#$%&()*+,-./:;<=>?@[/]^_`{|}~"""
    digits = '0123456789'
    ascii_letters = ascii_lowercase + ascii_uppercase
    return ''.join(sample(punctuation+ascii_letters+digits, 24))


def _pymongo_conf(dbname: str) -> str:
    return f"""
    "pymongo": {'{'}
        "url": "mongodb://localhost:27017/{dbname}"
    {'}'},""".strip('\n')

def _package_conf() -> str:
    return f"""
    "package": {'{'}
        "ts": {'{'}
            "url": "http://127.0.0.1:8000"
        {'}'},
        "swift": {'{'}
            "url": "http://127.0.0.1:8000"
        {'}'},
        "kotlin": {'{'}
            "url": "http://127.0.0.1:8000"
        {'}'}
    {'}'}""".strip('\n')



"""setup.py"""
from pathlib import Path
from setuptools import setup, find_packages
version = exec(open("jsonclasses_cli/version.py").read())
# The text of the README file
README = (Path(__file__).parent / "README.md").read_text()

setup(
    name='jsonclasses-cli',
    version=version,
    description=('JSONClasses CLI'),
    long_description=README,
    long_description_content_type='text/markdown',
    author='Fillmula Inc.',
    author_email='victor.teo@fillmula.com',
    license='MIT',
    packages=find_packages(exclude=('tests')),
    package_data={'jsonclasses_cli': ['py.typed']},
    zip_safe=False,
    url='https://github.com/fillmula/jsonclasses-cli',
    include_package_data=True,
    python_requires='>=3.10',
    install_requires=[
        'click>=8.0.3,<9.0.0',
        'rich>=10.12.0,<11.0.0',
        'jsonclasses>=3.3.0,<4.0.0',
        'jsonclasses-server>=3.3.0,<4.0.0'
    ],
    entry_points={
        'console_scripts': [
            'jsonclasses = jsonclasses_cli:app',
        ],
    },
)



from __future__ import annotations
from os import getcwd
from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from jsonclasses_cli.package import package


class TestPackageTs(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.temp_dir = TemporaryDirectory()
        cls.temp_path = Path(str(cls.temp_dir.name)) / "app_path"
        cls.cls_dir = Path(getcwd()) / 'tests' / 'classes'
        cls.data_dir = Path(getcwd()) / 'tests' / 'data_package_ts'
        cls.ts_path = cls.temp_path / 'packages' / 'ts'

    @classmethod
    def tearDownClass(cls) -> None:
        cls.temp_dir.cleanup()

    def test_package_create_all_files(self) -> None:
        package(self.temp_path, self.cls_dir / 'simple_song', 'ts', 'simple', True)
        gitignore_path = self.ts_path / '.gitignore'
        package_json_path = self.ts_path  / 'package.json'
        config_path = self.ts_path / 'tsconfig.json'
        api_path = self.ts_path  / 'src' / 'index.ts'
        self.assertTrue(gitignore_path.is_file(), gitignore_path.name)
        self.assertTrue(package_json_path.is_file(), package_json_path.name)
        self.assertTrue(config_path.is_file, config_path.name)
        self.assertTrue(api_path.is_file, api_path.name)

    def test_package_content_of_gitignore_packagejson_tsconfig(self) -> None:
        package(self.temp_path, self.cls_dir / 'simple_song', 'ts', 'simple', True)
        gitignore_path = self.ts_path / '.gitignore'
        test_gitignore_path = self.data_dir / '.gitignore'
        package_json_path =  self.ts_path  / 'package.json'
        test_package_json_path = self.data_dir / 'package.json'
        config_path =  self.ts_path / 'tsconfig.json'
        test_config_path = self.data_dir / 'tsconfig.json'
        self.assertEqual(gitignore_path.read_text(), test_gitignore_path.read_text())
        self.assertEqual(package_json_path.read_text(), test_package_json_path.read_text())
        self.assertEqual(config_path.read_text(), test_config_path.read_text())

    def test_package_create_without_link_and_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'simple_song.py', 'ts', 'simple', True)
        result = self.ts_path / 'src' / 'index.ts'
        expect = self.data_dir / 'simple_song_api.ts'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'session.py', 'ts', 'session', True)
        result = self.ts_path / 'src' / 'index.ts'
        expect = self.data_dir / 'session_api.ts'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkedthru_and_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkedthru_session.py', 'ts', 'linkedthru_session', True)
        result = self.ts_path / 'src' / 'index.ts'
        expect = self.data_dir / 'linkedthru_session_api.ts'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkedthru(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkedthru.py', 'ts', 'linkedthru', True)
        result = self.ts_path / 'src' / 'index.ts'
        expect = self.data_dir / 'linkedthru_api.ts'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkto_and_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkto_session.py', 'ts', 'linkto_session', True)
        result = self.ts_path / 'src' / 'index.ts'
        expect = self.data_dir / 'linkto_session_api.ts'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkto(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkto.py', 'ts', 'linkto', True)
        result = self.ts_path / 'src' / 'index.ts'
        expect = self.data_dir / 'linkto_api.ts'
        self.assertEqual(result.read_text(), expect.read_text())



from inflection import camelize
from jsonclasses.fdef import FType
from jsonclasses.cdef import CDef
from .interface import (
    interface, interface_first_line, interface_include_item, interface_include_key_item, interface_inst_items,
    interface_item, interface_pick_omit_items, interface_type_item, list_query_limit_skip_pn_ps, list_query_order_item)
from .shared_utils import (
    interface_required_include, field_ref_id_name, is_field_local_key, is_field_primary,
    is_field_ref, is_field_required_for_read, field_can_read, field_can_create, is_field_required_for_create,
    field_can_update, is_list_field, is_field_required_null_for_update, list_query_items, string,
    is_field_queryable)
from ...utils.package_utils import (
    to_create_input, to_include, to_include_key, to_list_query, to_result, to_result_picks, to_seek_query, to_single_query,
    to_sort_orders, to_update_input, to_query_data, is_field_link)
from ...utils.join_lines import join_lines
from .jtype_to_ts_type import jtype_to_ts_type


def data_interface(cdef: CDef) -> str:
    return join_lines([
        _interface_result(cdef),
        _interface_create_input(cdef),
        _interface_update_input(cdef),
        _interface_sort_order(cdef),
        _interface_result_pick(cdef),
        _interface_include_keys(cdef),
        _interface_include_type(cdef),
        _interface_single_query(cdef),
        _interface_list_query(cdef),
        _interface_seek_query(cdef),
        _interface_query_data(cdef),
    ], 2)


def _interface_result(cdef: CDef) -> str:
    items: list[str] = []
    for field in cdef.fields:
        if not field_can_read(field):
            continue
        name = camelize(field.name)
        ftype = jtype_to_ts_type(field.fdef, 'R')
        optional = not is_field_required_for_read(field)
        item = interface_item(name, ftype, optional)
        items.append(item)
        local_key = is_field_local_key(field)
        if local_key:
            is_list = '[]' if is_list_field(field) else ''
            idname = field_ref_id_name(field)
            idtype = jtype_to_ts_type(cdef.primary_field.fdef, 'R') + is_list
            item = interface_item(idname, idtype, optional)
            items.append(item)
    name = to_result(cdef)
    return interface(name, items, True)


def _interface_create_input(cdef: CDef) -> str:
    items: list[str] = []
    for field in cdef.fields:
        if not field_can_create(field):
            continue
        optional = not is_field_required_for_create(field)
        name = camelize(field.name)
        ftype = jtype_to_ts_type(field.fdef, 'C', is_field_link(field.fdef))
        optional = not is_field_required_for_read(field)
        item = interface_item(name, ftype, optional)
        items.append(item)
        local_key = is_field_local_key(field)
        if local_key:
            is_list = '[]' if is_list_field(field) else ''
            idname = field_ref_id_name(field)
            idtype = jtype_to_ts_type(cdef.primary_field.fdef, 'C') + is_list
            item = interface_item(idname, idtype, optional)
            items.append(item)
    name = to_create_input(cdef)
    return interface(name, items, True)


def _interface_update_input(cdef: CDef) -> str:
    items: list[str] = []
    for field in cdef.fields:
        if not field_can_update(field):
            continue
        null_for_update = '' if is_field_required_null_for_update(field) else ' | null'
        name = camelize(field.name)
        ftype = jtype_to_ts_type(field.fdef, 'U', is_field_link(field.fdef)) + null_for_update
        item = interface_item(name, ftype, True)
        items.append(item)
        local_key = is_field_local_key(field)
        if local_key:
            is_list = '[]' if is_list_field(field) else ''
            idname = field_ref_id_name(field)
            idtype = jtype_to_ts_type(cdef.primary_field.fdef, 'U') + is_list + null_for_update
            item = interface_item(idname, idtype, True)
            items.append(item)
    name = to_update_input(cdef)
    return interface(name, items, True)


def _interface_sort_order(cdef: CDef) -> str:
    items: list[str] = []
    for field in cdef.fields:
        if not is_field_queryable(field):
            continue
        if is_field_primary(field):
            continue
        if is_field_ref(field):
            continue
        if not field_can_read(field):
            continue
        name = camelize(field.name)
        items.append(string(name))
        items.append(string('-' + name))
    name = to_sort_orders(cdef)
    return interface_type_item(name, items)


def _interface_result_pick(cdef: CDef) -> str:
    items: list[str] = []
    for field in cdef.fields:
        if not field_can_read(field):
            continue
        name = string(camelize(field.name))
        items.append(name)
        if is_field_local_key(field):
            idname = string(field_ref_id_name(field))
            items.append(idname)
    name = to_result_picks(cdef)
    return interface_type_item(name, items)


def _interface_include_keys(cdef: CDef) -> str:
    cname = cdef.name
    keys: list[str] = []
    for field in cdef.fields:
        if is_field_ref(field):
            if field.fdef.ftype == FType.LIST:
                ftype = jtype_to_ts_type(field.fdef, 'R').removesuffix('[]')
                include_type = ftype + 'ListQuery'
            else:
                ftype = jtype_to_ts_type(field.fdef, 'R')
                include_type = ftype + 'SingleQuery'
            name = to_include_key(cname, field.name)
            keys.append(_interface_include_key(name, field.name, include_type))
    return join_lines(keys, 2)


def _interface_include_key(name: str, key: str, ftype: str) -> str:
    return join_lines([
        interface_first_line(name),
        interface_include_key_item(key, ftype),
        '}'
    ])


def _interface_include_type(cdef: CDef) -> str:
    cname = cdef.name
    include = to_include(cdef)
    include_types: list[str] = []
    for field in cdef.fields:
        if is_field_ref(field):
            name = to_include_key(cname, field.name)
            include_types.append(name)
    return interface_type_item(include, include_types) if len(include_types) else ""


def _interface_single_query(cdef: CDef) -> str:
    name = to_single_query(cdef)
    result_pick_name = to_result_picks(cdef)
    return join_lines([
        interface_first_line(name),
        interface_pick_omit_items(result_pick_name),
        _single_query_include(cdef),
        '}'
    ])


def _single_query_include(cdef: CDef) -> str:
    name = to_include(cdef)
    return interface_include_item(name) if interface_required_include(cdef) else ""


def _interface_list_query(cdef: CDef) -> str:
    name = to_list_query(cdef)
    items = list(map(lambda i: interface_item(i[0], i[1], True), list_query_items(cdef)))
    order = to_sort_orders(cdef)
    pick = to_result_picks(cdef)
    return join_lines([
        interface_first_line(name),
        interface_inst_items(items),
        list_query_order_item(order),
        list_query_limit_skip_pn_ps(),
        interface_pick_omit_items(pick),
        _single_query_include(cdef),
        '}'
    ])


def _interface_seek_query(cdef: CDef) -> str:
    name = to_seek_query(cdef)
    items = list(map(lambda i: interface_item(i[0], i[1], True), list_query_items(cdef)))
    return join_lines([
        interface_first_line(name),
        interface_inst_items(items),
        '}'
    ])


def _interface_query_data(cdef: CDef) -> str:
    name = to_query_data(cdef)
    items = [
        interface_item('_query', to_seek_query(cdef), False),
        interface_item('_data', to_update_input(cdef), False)
    ]
    return join_lines([
        interface_first_line(name),
        interface_inst_items(items),
        '}'
    ])



def number_query() -> str:
    return """
interface NumberValueQuery {
    _gt?: number
    _gte?: number
    _lt?: number
    _lte?: number
}

interface NumberEqQuery {
    _eq: number
}

interface NumberNeqQuery {
    _neq: number
}

interface NumberNullQuery {
    _null: boolean
}

interface NumberOrQuery {
    _or: NumberQuery[]
}

interface NumberAndQuery {
    _and: NumberQuery[]
}

export type NumberQuery = number | NumberEqQuery | NumberNeqQuery | NumberNullQuery | NumberValueQuery | NumberOrQuery | NumberAndQuery
    """.strip() + '\n'





def session_manager(items: dict[str, str]) -> str:
    session_items: list[str] = []
    for (_, c) in items.items():
        session_items.append(c + 'Session')
    sessions = ' | '.join(session_items)
    return f"""
class SessionManager {'{'}

    #sessionKey = '_jsonclasses_session'
    #session: {sessions} | undefined

    static share = new SessionManager()

    constructor() {'{'}
        const item = localStorage.getItem(this.#sessionKey)
        if (item && item !== null && item !== '') {'{'}
            this.#session = JSON.parse(item)
        {'}'} else {'{'}
            this.#session = undefined
        {'}'}
    {'}'}

    setSession(session: {sessions} | undefined | null) {'{'}
        if (session) {'{'}
            this.#session = session
            localStorage.setItem(this.#sessionKey, JSON.stringify(session))
        {'}'} else {'{'}
            this.#session = undefined
            localStorage.removeItem(this.#sessionKey)
        {'}'}
    {'}'}

    hasSession(): boolean {'{'}
        return this.#session !== undefined
    {'}'}

    getToken(): string | undefined {'{'}
        return this.#session?.token
    {'}'}

    getSession(): {sessions} | undefined {'{'}
        return this.#session
    {'}'}

    clearSession() {'{'}
        this.#session = undefined
        localStorage.removeItem(this.#sessionKey)
    {'}'}
{'}'}
    """.strip() + "\n"



from pathlib import Path
from inflection import underscore, dasherize


def package_json_content(dest: Path):
    pkg_name = dasherize(underscore(dest.name))
    return f"""
{'{'}
    "name": "{pkg_name}",
    "version": "0.1.0",
    "private": true,
    "description": "This API client package is generated by JSONClasses CLI.",
    "main": "lib/index.js",
    "types": "src/index.ts",
    "author": "",
    "dependencies": {'{'}
        "axios": "^0.24.0",
        "qsparser-js": "^1.0.1"
    {'}'},
    "devDependencies": {'{'}
        "typescript": "^4.4.4"
    {'}'}
{'}'}
    """.strip() + '\n'



from __future__ import annotations
from os import getcwd
from unittest import TestCase
from tempfile import TemporaryDirectory
from pathlib import Path
from jsonclasses_cli.package import package


class TestPackageSwift(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.temp_dir = TemporaryDirectory()
        cls.temp_path = Path(str(cls.temp_dir.name)) / "swift_path"
        cls.cls_dir = Path(getcwd()) / 'tests' / 'classes'
        cls.data_path = Path(getcwd()) / 'tests' / 'data_package_swift'
        cls.swift_path = cls.temp_path / 'packages' / 'swift'

    @classmethod
    def tearDownClass(cls) -> None:
        cls.temp_dir.cleanup()

    def test_package_swift_create_all_files(self) -> None:
        package(self.temp_path, self.cls_dir / 'simple_song', 'swift', 'simple', True)
        gitignore = self.swift_path / '.gitignore'
        package_swift = self.swift_path  / 'Package.swift'
        read_me = self.swift_path  / 'README.md'
        api = self.swift_path  / 'Sources' / 'API' / 'API.swift'
        self.assertTrue(gitignore.is_file(), gitignore.name)
        self.assertTrue(package_swift.is_file(), package_swift.name)
        self.assertTrue(read_me.is_file(), read_me.name)
        self.assertTrue(api.is_file(), api.name)

    def test_package_swift_content_of_gitignore_package_and_readme(self) -> None:
        package(self.temp_path, self.cls_dir / 'simple_song', 'swift', 'simple', True)
        package_swift = self.swift_path / 'Package.swift'
        gitignore = self.swift_path / '.gitignore'
        read_me = self.swift_path / 'README.md'
        expect_package_swift = self.data_path / 'Package.swift'
        expect_gitignore = self.data_path / '.gitignore'
        expect_read_me = self.data_path / 'README.md'
        self.assertEqual(package_swift.read_text(), expect_package_swift.read_text())
        self.assertEqual(gitignore.read_text(), expect_gitignore.read_text())
        self.assertEqual(read_me.read_text(), expect_read_me.read_text())

    def test_package_create_without_link_and_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'simple_song', 'swift', 'simple', True)
        result = self.swift_path / 'Sources' / 'API' / 'API.swift'
        expect = self.data_path / 'simple_api.swift'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'session.py', 'swift', 'session', True)
        result = self.swift_path / 'Sources' / 'API' / 'API.swift'
        expect = self.data_path / 'session_api.swift'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkedthru_and_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkedthru_session.py', 'swift', 'linkedthru_session', True)
        result = self.swift_path / 'Sources' / 'API' / 'API.swift'
        expect = self.data_path / 'linkedthru_session_api.swift'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkedthru(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkedthru.py', 'swift', 'linkedthru', True)
        result = self.swift_path / 'Sources' / 'API' / 'API.swift'
        expect = self.data_path / 'linkedthru_api.swift'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkto_and_session(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkto_session.py', 'swift', 'linkto_session', True)
        result = self.swift_path / 'Sources' / 'API' / 'API.swift'
        expect = self.data_path / 'linkto_session_api.swift'
        self.assertEqual(result.read_text(), expect.read_text())

    def test_package_create_with_linkto(self) -> None:
        package(self.temp_path, self.cls_dir / 'linkto.py', 'swift', 'linkto', True)
        result = self.swift_path / 'Sources' / 'API' / 'API.swift'
        expect = self.data_path / 'linkto_api.swift'
        self.assertEqual(result.read_text(), expect.read_text())



from inflection import camelize
from jsonclasses.cdef import CDef
from jsonclasses.fdef import FDef
from jsonclasses.jfield import JField
from jsonclasses.fdef import (
    FStore, FType, Nullability, ReadRule, Queryability, WriteRule
)
from jsonclasses.modifiers.required_modifier import RequiredModifier
from jsonclasses.modifiers.default_modifier import DefaultModifier
from .jtype_to_ts_type import jtype_to_ts_type


def list_query_items(cdef: CDef) -> list[tuple[str, str]]:
    items: list[tuple[str, str]] = []
    for field in cdef.fields:
        if not is_field_queryable(field):
            continue
        name = camelize(field.name)
        type = jtype_to_ts_type(field.fdef, 'Q')
        if is_field_ref(field):
            if not is_field_local_key(field):
                continue
            idname = field_ref_id_name(field)
            items.append((idname, 'IDQuery'))
        else:
            items.append((name, type))
    return items


def is_field_local_key(field: JField) -> bool:
    return field.fdef.fstore == FStore.LOCAL_KEY



def is_field_required_for_create(field: JField) -> bool:
    if field_has_default(field):
        return False
    if is_field_nonnull(field):
        return True
    return next((True for v in field.types.modifier.vs if isinstance(v, RequiredModifier)), False)


def is_field_required_null_for_update(field: JField):
    if is_field_nonnull(field):
        return True
    return next((True for v in field.types.modifier.vs if isinstance(v, RequiredModifier)), False)


def field_ref_id_name(field: JField) -> str:
    rkes = field.cdef.jconf.ref_name_strategy
    kes = field.cdef.jconf.input_key_strategy
    return kes(rkes(field))


def field_has_default(field: JField) -> bool:
    if is_field_nonnull(field):
        return True
    return next((True for v in field.types.modifier.vs if isinstance(v, DefaultModifier)), False)


def is_field_nonnull(field: JField) -> bool:
    if field.fdef.ftype == FType.LIST:
        if field.fdef.fstore == FStore.LOCAL_KEY or field.fdef.fstore == FStore.FOREIGN_KEY:
            if field.fdef.collection_nullability == Nullability.NONNULL:
                return True
    return False


def is_field_primary(field: JField) -> bool:
    return field.fdef.primary


def is_field_ref(field: JField) -> bool:
    if field.fdef.fstore == FStore.LOCAL_KEY:
        return True
    if field.fdef.fstore == FStore.FOREIGN_KEY:
        return True
    return False


def is_field_required_for_read(field: JField) -> bool:
    if is_field_nonnull(field):
        return True
    return next((True for v in field.types.modifier.vs if isinstance(v, RequiredModifier)), False)


def is_field_queryable(field: JField) -> bool:
    if field.fdef.read_rule == ReadRule.NO_READ:
        return False
    if field.fdef.fstore == FStore.TEMP:
        return False
    return field.fdef.queryability != Queryability.UNQUERYABLE


def field_can_create(field: JField) -> bool:
    return field.fdef.write_rule != WriteRule.NO_WRITE


def field_can_update(field: JField) -> bool:
    if field.fdef.write_rule == WriteRule.NO_WRITE:
        return False
    if field.fdef.write_rule == WriteRule.WRITE_ONCE:
        if is_field_required_for_create(field):
            return False
    return True


def field_can_read(field: JField) -> bool:
    if field.fdef.read_rule == ReadRule.NO_READ:
        return False
    if field.fdef.fstore == FStore.TEMP:
        return False
    return True


def interface_required_include(cdef: CDef) -> bool:
    items = [f for f in cdef.fields if is_field_ref(f)]
    return len(items) > 0


def is_list_field(field: JField) -> bool:
    return field.fdef.ftype == FType.LIST


def string(val: str) -> str:
    return "'" +val +"'"



from typing import Literal
from ...utils.join_lines import join_lines


CodableClassItem = tuple[
    Literal['public', 'private', 'fileprivate'],
    Literal['var', 'let'],
    str,
    str,
    bool
]


def codable_class(name: str, items: list[CodableClassItem], unwrapped: bool = False) -> str:
    return join_lines([
        join_lines([
            _codable_class_first_line(name),
            _codable_class_inst_vars(items, unwrapped),
        ], 1),
        join_lines([
            '    public init(',
            _codable_class_init_param_list(items),
            '    ) {',
            _codable_class_init_assigns(items),
            '    }',
            '}'
        ], 1)
    ], 2)


def _codable_class_first_line(name: str) -> str:
    return f"public class {name}: Codable {'{'}"


def _codable_class_inst_vars(items: list[CodableClassItem], unwrapped: bool) -> str:
    return join_lines(map(lambda i: f"    {i[0]} {i[1]} {i[2]}: {i[3]}{'!' if unwrapped else '?' if i[4] else ''}", items), 1)


def _codable_class_init_param_list(items: list[CodableClassItem]) -> str:
    last = len(items) - 1
    return join_lines(map(lambda i: f"        {i[1][2]}: {i[1][3]}{'? = nil' if i[1][4] else ''}{'' if i[0] == last else ','}", enumerate(items)), 1)


def _codable_class_init_assigns(items: list[CodableClassItem]) -> str:
    return join_lines(map(lambda i: f"        self.{i[2]} = {i[2]}", items), 1)


def codable_class_item(access: Literal["public", "fileprivate", "private"],
                       mutability: Literal["let", "var"],
                       name: str,
                       field_type: str,
                       optional: bool) -> CodableClassItem:
    return (access, mutability, name, field_type, optional)
